@@ -218,7 +218,9 @@
                   old-context old-length
                   context length)
             t))
-        (let ((old-bitmap (wrapper-bitmap old-layout)))
+        ;; The "%" accessor reads the bitmap from the layout,
+        ;; not from its related defstruct-description.
+        (let ((old-bitmap (%layout-bitmap old-layout)))
           (unless (= old-bitmap bitmap)
             (warn "change in placement of raw slots of class ~S ~
 between the ~A definition and the ~A definition"
@@ -230,35 +232,43 @@ between the ~A definition and the ~A definition"
                 name old-context context)
           t))))
 
+;;; Used by the loader to forward-reference layouts for classes whose
+;;; definitions may not have been loaded yet. This allows type tests
+;;; to be loaded when the type definition hasn't been loaded yet.
+;;;
+;;; If we can't find any existing layout, then we create a new one
+;;; with the supplied information, storing it in
+;;; *FORWARD-REFERENCED-LAYOUTS*. If we can find the layout, then
+;;; return it, after checking for compatibility. If incompatible, we
+;;; allow the layout to be replaced, altered or left alone.
 (defun load-layout (name depthoid inherits length bitmap flags)
-  (let* ((layout
-          (or (binding* ((classoid (find-classoid name nil) :exit-if-null))
-                (classoid-wrapper classoid))
-              (let ((table *forward-referenced-wrappers*))
-                (with-world-lock ()
+  (let* ((table *forward-referenced-wrappers*)
+         (classoid (find-classoid name nil))  ; thread safety
+         (new-layout (make-layout (hash-layout-name name)
+                                  (or classoid (make-undefined-classoid name))
+                                  :depthoid depthoid :inherits inherits
+                                  :length length :bitmap bitmap :flags flags))
+         (existing-layout
+           (or (and classoid (classoid-wrapper classoid))
+               (with-world-lock ()
                  (let ((classoid (find-classoid name nil)))
+                   (when classoid
+                     (setf (wrapper-classoid new-layout) classoid))
                    (or (and classoid (classoid-wrapper classoid))
-                       (ensure-gethash
-                        name table
-                        (make-layout
-                         (hash-layout-name name)
-                         (or classoid (make-undefined-classoid name))
-                         :depthoid depthoid :inherits inherits
-                         :length length :bitmap bitmap :flags flags))))))))
+                       (ensure-gethash name table new-layout))))))
          (classoid
-          (or (find-classoid name nil) (wrapper-classoid layout))))
-    (if (or (eq (wrapper-invalid layout) :uninitialized)
-            (not *type-system-initialized*))
-        (setf (wrapper-classoid layout) classoid)
-        ;; There was an old layout already initialized with old
-        ;; information, and we'll now check that old information
-        ;; which was known with certainty is consistent with current
-        ;; information which is known with certainty.
-        (when (warn-if-altered-layout "current" layout "compile time"
-                                    length inherits depthoid bitmap)
-          (error "The loaded code expects an incompatible layout for class ~S."
-                 (wrapper-proper-name layout))))
-    layout))
+           (or (find-classoid name nil) (wrapper-classoid existing-layout))))
+    (cond ((or (eq (wrapper-invalid existing-layout) :uninitialized)
+               (not *type-system-initialized*))
+           (setf (wrapper-classoid existing-layout) classoid))
+          ;; There was an old layout already initialized with old
+          ;; information, and we'll now check that old information
+          ;; which was known with certainty is consistent with current
+          ;; information which is known with certainty.
+          ((warn-if-altered-layout "current" existing-layout "compile time"
+                                   length inherits depthoid bitmap)
+           (%redefine-defstruct classoid existing-layout new-layout)))
+    existing-layout))
 
 (defun classoid-lock (classoid)
   #+sb-xc-host (declare (ignore classoid))
@@ -269,10 +279,13 @@ between the ~A definition and the ~A definition"
         (if (eq oldval nil) lock oldval))))
 
 (defun add-subclassoid (super sub wrapper)
+  (declare (sb-c::tlab :system))
   (with-system-mutex ((classoid-lock super))
     (let ((table (classoid-subclasses super))
           (count 0))
       (cond ((hash-table-p table)
+             ;; If the table needs to resize, it'll stay on the heap
+             ;; even if we're using an arena.
              (setf (gethash sub table) wrapper))
             ((dolist (cell table)
                (when (eq (car cell) sub)
@@ -288,8 +301,10 @@ between the ~A definition and the ~A definition"
             (t
              ;; Upgrade to a hash-table
              (let ((new #+sb-xc-host (make-hash-table :test 'eq)
-                        #-sb-xc-host (make-hash-table :hash-function #'type-hash-value
-                                                      :test 'eq)))
+                        #-sb-xc-host
+                        (sb-vm:without-arena "add-subclassoid"
+                            (make-hash-table :hash-function #'type-hash-value
+                                             :test 'eq))))
                (loop for (key . val) in table do (setf (gethash key new) val))
                (setf (gethash sub new) wrapper)
                (setf (classoid-subclasses super) new)
@@ -343,22 +358,18 @@ between the ~A definition and the ~A definition"
                table))
   nil)
 
-;;; Record LAYOUT as the layout for its class, adding it as a subtype
+;;; Record WRAPPER as the layout for its class, adding it as a subtype
 ;;; of all superclasses. This is the operation that "installs" a
 ;;; layout for a class in the type system, clobbering any old layout.
 ;;; However, this does not modify the class namespace; that is a
 ;;; separate operation (think anonymous classes.)
 ;;; -- If INVALIDATE, then all the layouts for any old definition
 ;;;    and subclasses are invalidated, and the SUBCLASSES slot is cleared.
-;;; -- If DESTRUCT-LAYOUT, then this is some old layout, and is to be
-;;;    destructively modified to hold the same type information.
+;;; -- If MODIFY is given, then it is some old layout, and is to be
+;;;    destructively altered to hold the same data as WRAPPER.
 (macrolet ((set-bitmap-from-layout (to-layout from-layout)
-             `(let ((to-index
-                     (+ (type-dd-length sb-vm:layout)
-                        (calculate-extra-id-words (layout-depthoid ,to-layout))))
-                    (from-index
-                     (+ (type-dd-length sb-vm:layout)
-                        (calculate-extra-id-words (layout-depthoid ,from-layout)))))
+             `(let ((to-index (bitmap-start ,to-layout))
+                    (from-index (bitmap-start ,from-layout)))
                 (dotimes (i (bitmap-nwords ,from-layout))
                   (%raw-instance-set/word ,to-layout (+ to-index i)
                         (%raw-instance-ref/word ,from-layout (+ from-index i)))))))
@@ -468,6 +479,7 @@ between the ~A definition and the ~A definition"
 ;;; not be read as being in CPL order.
 (defun order-layout-inherits (layouts)
   (declare (simple-vector layouts))
+  (declare (sb-c::tlab :system))
   (let ((length (length layouts))
         (max-depth -1))
     (dotimes (i length)
@@ -716,11 +728,10 @@ between the ~A definition and the ~A definition"
        (values-specifier-type-cache-clear)))))
 
 (defun find-classoid-cell (name &key create)
-  (let ((real-name (uncross name)))
-    (cond ((info :type :classoid-cell real-name))
-          (create
-           (get-info-value-initializing :type :classoid-cell real-name
-                                        (make-classoid-cell real-name))))))
+  (cond ((info :type :classoid-cell name))
+        (create
+         (get-info-value-initializing :type :classoid-cell name
+                                      (make-classoid-cell name)))))
 
 ;;; Return the classoid with the specified NAME. If ERRORP is false,
 ;;; then NIL is returned when no such class exists.
@@ -924,7 +935,7 @@ between the ~A definition and the ~A definition"
 (declaim (type cons **non-instance-classoid-types**))
 (defglobal **non-instance-classoid-types**
   '(symbol system-area-pointer weak-pointer code-component
-    #-(or x86 x86-64 arm64) lra
+    #-(or x86 x86-64 arm64 riscv) lra
     fdefn random-class))
 
 (defun classoid-non-instance-p (classoid)
@@ -989,270 +1000,300 @@ between the ~A definition and the ~A definition"
 ;;; (note the difference in spelling, to help keep things unconfusing)
 #+sb-xc-host
 (defvar *builtin-classoids*
-   `((t :state :read-only :translation t)
-     (character :codes (,sb-vm:character-widetag)
-                :translation (character-set)
-                :prototype-form (code-char 42))
-     (symbol :codes (,sb-vm:symbol-widetag)
-             :predicate symbolp
-             :prototype-form '*)
+  ;; This variable holds data expressed as conses, some of which are "similar" (in
+  ;; this case, EQUAL) to each other, but not EQL to each other at read-time.
+  ;; Host compilers can differ in how they treat "similar" data on
+  ;; file-compilation; they are entitled to coalesce data to be EQL.  Differences
+  ;; in those host choices then manifest (cosmetically) in target data, where the
+  ;; resulting value of *BUILTIN-CLASSOIDS* at cold-load time differs in its
+  ;; structure between e.g. SBCL and CCL hosts.
+  ;;
+  ;; To avoid this difference, we postprocess the data structure to
+  ;; convert all similar data within it to be EQL, at which point the
+  ;; host compiler no longer has choices: EQL data must remain EQL
+  ;; after file-compilation.
+  ;;
+  ;; FIXME: it might be nice if our cross-compile backquote reader
+  ;; did hash-consing for us automatically.  Without thinking too
+  ;; hard about it, though, it might need to be sufficiently smart
+  ;; to be able to handle nested backquotations and/or
+  ;; backquotations within unquotations.
+  (let ((table (make-hash-table :test 'equal)))
+    (labels ((hash-cons (thing)
+               (cond
+                 ((atom thing) thing)
+                 ((gethash thing table))
+                 (t (setf (gethash thing table)
+                          (cons (hash-cons (car thing)) (hash-cons (cdr thing))))))))
+      (hash-cons
+       `((t :state :read-only :translation t)
+         (character :codes (,sb-vm:character-widetag)
+                    :translation (character-set)
+                    :prototype-form (code-char 42))
+         (symbol :codes (,sb-vm:symbol-widetag)
+                 :predicate symbolp
+                 :prototype-form '*)
 
-     (system-area-pointer :codes (,sb-vm:sap-widetag)
-                          :predicate system-area-pointer-p
-                          :prototype-form (int-sap 0))
-     (weak-pointer :codes (,sb-vm:weak-pointer-widetag)
-                   :predicate weak-pointer-p
-                   :prototype-form (make-weak-pointer 0))
-     (code-component :codes (,sb-vm:code-header-widetag)
-                     :predicate code-component-p
-                     :prototype-form (fun-code-header #'identity))
-     #-(or x86 x86-64 arm64) (lra :codes (,sb-vm:return-pc-widetag)
-                                  :predicate lra-p
-                                  ;; Make the PROTOTYPE slot unbound.
-                                  :prototype-form sb-pcl:+slot-unbound+)
-     (fdefn :codes (,sb-vm:fdefn-widetag)
-            :predicate fdefn-p
-            :prototype-form (find-or-create-fdefn 'sb-mop:class-prototype))
-     (random-class ; used for unknown type codes
-            ;; Make the PROTOTYPE slot unbound.
-            :prototype-form sb-pcl:+slot-unbound+)
-     (function
-      :codes (,sb-vm:closure-widetag ,sb-vm:simple-fun-widetag)
-      :predicate functionp
-      :state :read-only
-      :prototype-form #'identity)
+         (system-area-pointer :codes (,sb-vm:sap-widetag)
+                              :predicate system-area-pointer-p
+                              :prototype-form (int-sap 0))
+         (weak-pointer :codes (,sb-vm:weak-pointer-widetag)
+                       :predicate weak-pointer-p
+                       :prototype-form (make-weak-pointer 0))
+         (code-component :codes (,sb-vm:code-header-widetag)
+                         :predicate code-component-p
+                         :prototype-form (fun-code-header #'identity))
+         #-(or x86 x86-64 arm64 riscv)
+         (lra :codes (,sb-vm:return-pc-widetag)
+              :predicate lra-p
+              ;; Make the PROTOTYPE slot unbound.
+              :prototype-form sb-pcl:+slot-unbound+)
+         (fdefn :codes (,sb-vm:fdefn-widetag)
+                :predicate fdefn-p
+                :prototype-form (find-or-create-fdefn '(setf car)))
+         (random-class             ; used for unknown type codes
+          ;; Make the PROTOTYPE slot unbound.
+          :prototype-form sb-pcl:+slot-unbound+)
+         (function
+          :codes (,sb-vm:closure-widetag ,sb-vm:simple-fun-widetag)
+          :predicate functionp
+          :state :read-only
+          :prototype-form #'identity)
 
-     (number :translation number :prototype-form 0)
-     (complex
-      :translation complex
-      :inherits (number)
-      :codes (,sb-vm:complex-widetag)
-      :prototype-form ,(complex 0 1))
-     (complex-single-float
-      :translation (complex single-float)
-      :inherits (complex number)
-      :codes (,sb-vm:complex-single-float-widetag)
-      :prototype-form ,(complex $0f0 $0f0))
-     (complex-double-float
-      :translation (complex double-float)
-      :inherits (complex number)
-      :codes (,sb-vm:complex-double-float-widetag)
-      :prototype-form ,(complex $0d0 $0d0))
-     #+long-float
-     (complex-long-float
-      :translation (complex long-float)
-      :inherits (complex number)
-      :codes (,sb-vm:complex-long-float-widetag)
-      :prototype-form ,(complex $0L0 $0L0))
-     #+sb-simd-pack
-     (simd-pack
-      :translation simd-pack
-      :codes (,sb-vm:simd-pack-widetag)
-      :prototype-form (%make-simd-pack-ub64 42 42))
-     #+sb-simd-pack-256
-     (simd-pack-256
-      :translation simd-pack-256
-      :codes (,sb-vm:simd-pack-256-widetag)
-      :prototype-form
-      ;; KLUDGE: doesn't work without AVX2 support from the CPU
-      ;; (%make-simd-pack-256-ub64 42 42 42 42)
-      sb-pcl:+slot-unbound+)
-     (real :translation real :inherits (number) :prototype-form 0)
-     (float :translation float :inherits (real number) :prototype-form $0f0)
-     (single-float
-      :translation single-float
-      :inherits (float real number)
-      :codes (,sb-vm:single-float-widetag)
-      :prototype-form $0f0)
-     (double-float
-      :translation double-float
-      :inherits (float real number)
-      :codes (,sb-vm:double-float-widetag)
-      :prototype-form $0d0)
-     #+long-float
-     (long-float
-      :translation long-float
-      :inherits (float real number)
-      :codes (,sb-vm:long-float-widetag)
-      :prototype-form $0L0)
-     (rational
-      :translation rational :inherits (real number) :prototype-form 0)
-     (ratio
-      :translation (and rational (not integer))
-      :inherits (rational real number)
-      :codes (,sb-vm:ratio-widetag)
-      :prototype-form 1/42)
-     (integer
-      :translation integer :inherits (rational real number) :prototype-form 0)
-     (fixnum
-      :translation (integer ,most-negative-fixnum ,most-positive-fixnum)
-      :inherits (integer rational real number)
-      :codes ,(mapcar #'symbol-value sb-vm::fixnum-lowtags)
-      :prototype-form 42)
-     (bignum
-      :translation (and integer (not fixnum))
-      :inherits (integer rational real number)
-      :codes (,sb-vm:bignum-widetag)
-      :prototype-form ,(1+ most-positive-fixnum))
+         (number :translation number :prototype-form 0)
+         (complex
+          :translation complex
+          :inherits (number)
+          :codes (,sb-vm:complex-widetag)
+          :prototype-form ,(complex 0 1))
+         (complex-single-float
+          :translation (complex single-float)
+          :inherits (complex number)
+          :codes (,sb-vm:complex-single-float-widetag)
+          :prototype-form ,(complex $0f0 $0f0))
+         (complex-double-float
+          :translation (complex double-float)
+          :inherits (complex number)
+          :codes (,sb-vm:complex-double-float-widetag)
+          :prototype-form ,(complex $0d0 $0d0))
+         #+long-float
+         (complex-long-float
+          :translation (complex long-float)
+          :inherits (complex number)
+          :codes (,sb-vm:complex-long-float-widetag)
+          :prototype-form ,(complex $0L0 $0L0))
+         #+sb-simd-pack
+         (simd-pack
+          :translation simd-pack
+          :codes (,sb-vm:simd-pack-widetag)
+          :prototype-form (%make-simd-pack-ub64 42 42))
+         #+sb-simd-pack-256
+         (simd-pack-256
+          :translation simd-pack-256
+          :codes (,sb-vm:simd-pack-256-widetag)
+          :prototype-form
+          ;; KLUDGE: doesn't work without AVX2 support from the CPU
+          ;; (%make-simd-pack-256-ub64 42 42 42 42)
+          sb-pcl:+slot-unbound+)
+         (real :translation real :inherits (number) :prototype-form 0)
+         (float :translation float :inherits (real number) :prototype-form $0f0)
+         (single-float
+          :translation single-float
+          :inherits (float real number)
+          :codes (,sb-vm:single-float-widetag)
+          :prototype-form $0f0)
+         (double-float
+          :translation double-float
+          :inherits (float real number)
+          :codes (,sb-vm:double-float-widetag)
+          :prototype-form $0d0)
+         #+long-float
+         (long-float
+          :translation long-float
+          :inherits (float real number)
+          :codes (,sb-vm:long-float-widetag)
+          :prototype-form $0L0)
+         (rational
+          :translation rational :inherits (real number) :prototype-form 0)
+         (ratio
+          :translation (and rational (not integer))
+          :inherits (rational real number)
+          :codes (,sb-vm:ratio-widetag)
+          :prototype-form 1/42)
+         (integer
+          :translation integer :inherits (rational real number) :prototype-form 0)
+         (fixnum
+          :translation (integer ,most-negative-fixnum ,most-positive-fixnum)
+          :inherits (integer rational real number)
+          :codes ,sb-vm::fixnum-lowtags
+          :prototype-form 42)
+         (bignum
+          :translation (and integer (not fixnum))
+          :inherits (integer rational real number)
+          :codes (,sb-vm:bignum-widetag)
+          :prototype-form ,(1+ most-positive-fixnum))
 
-     (array :translation array :codes (,sb-vm:complex-array-widetag)
-            :hierarchical-p nil
-            :prototype-form (make-array nil :adjustable t))
-     (simple-array
-      :translation simple-array :codes (,sb-vm:simple-array-widetag)
-      :inherits (array)
-      :prototype-form (make-array nil))
-     (sequence
-      :translation (or cons (member nil) vector extended-sequence)
-      :state :read-only
-      :depth 1)
-     (vector
-      :translation vector :codes (,sb-vm:complex-vector-widetag)
-      :direct-superclasses (array sequence)
-      :inherits (array sequence)
-      :prototype-form (make-array 0 :adjustable t))
-     (simple-vector
-      :translation simple-vector :codes (,sb-vm:simple-vector-widetag)
-      :direct-superclasses (vector simple-array)
-      :inherits (vector simple-array array sequence)
-      :prototype-form (make-array 0))
-     (bit-vector
-      :translation bit-vector :codes (,sb-vm:complex-bit-vector-widetag)
-      :inherits (vector array sequence)
-      :prototype-form (make-array 0 :element-type 'bit :fill-pointer t))
-     (simple-bit-vector
-      :translation simple-bit-vector :codes (,sb-vm:simple-bit-vector-widetag)
-      :direct-superclasses (bit-vector simple-array)
-      :inherits (bit-vector vector simple-array
-                 array sequence)
-      :prototype-form #*)
-     (string
-      :translation string
-      :direct-superclasses (vector)
-      :inherits (vector array sequence)
-      :prototype-form "")
-     (simple-string
-      :translation simple-string
-      :direct-superclasses (string simple-array)
-      :inherits (string vector simple-array array sequence)
-      :prototype-form "")
-     (vector-nil
-      :translation (vector nil)
-      :inherits (vector array sequence)
-      :prototype-form (make-array 0 :element-type 'nil :fill-pointer t))
-     ;; This name is imperfect. It should be SIMPLE-RANK1-ARRAY-NIL
-     ;; to clearly convey that the dimensions are '(*) and not '*.
-     (simple-array-nil
-      :translation (simple-array nil (*))
-      :codes (,sb-vm:simple-array-nil-widetag)
-      :direct-superclasses (vector-nil)
-      :inherits (vector-nil vector simple-array array sequence)
-      :prototype-form (make-array 0 :element-type 'nil))
-     (base-string
-      :translation base-string
-      :codes (,sb-vm:complex-base-string-widetag)
-      :direct-superclasses (string)
-      :inherits (string vector array sequence)
-      :prototype-form (make-array 0 :element-type 'base-char :fill-pointer t))
-     (simple-base-string
-      :translation simple-base-string
-      :codes (,sb-vm:simple-base-string-widetag)
-      :direct-superclasses (base-string simple-string)
-      :inherits (base-string simple-string string vector simple-array
-                 array sequence)
-      :prototype-form (make-array 0 :element-type 'base-char))
-     #+sb-unicode
-     (character-string
-      :translation (vector character)
-      :codes (,sb-vm:complex-character-string-widetag)
-      :direct-superclasses (string)
-      :inherits (string vector array sequence)
-      :prototype-form (make-array 0 :element-type 'character :fill-pointer t))
-     #+sb-unicode
-     (simple-character-string
-      :translation (simple-array character (*))
-      :codes (,sb-vm:simple-character-string-widetag)
-      :direct-superclasses (character-string simple-string)
-      :inherits (character-string simple-string string vector simple-array
-                 array sequence)
-      :prototype-form (make-array 0 :element-type 'character))
-     (list
-      :translation (or cons (member nil))
-      :inherits (sequence)
-      :prototype-form 'nil)
-     (cons
-      :codes (,sb-vm:list-pointer-lowtag)
-      :translation cons
-      :inherits (list sequence)
-      :prototype-form (cons nil nil))
-     (null
-      :translation (member nil)
-      :inherits (symbol list sequence)
-      :direct-superclasses (symbol list)
-      :prototype-form 'nil)
+         (array :translation array :codes (,sb-vm:complex-array-widetag)
+                :hierarchical-p nil
+                :prototype-form (make-array nil :adjustable t))
+         (simple-array
+          :translation simple-array :codes (,sb-vm:simple-array-widetag)
+          :inherits (array)
+          :prototype-form (make-array nil))
+         (sequence
+          :translation (or cons (member nil) vector extended-sequence)
+          :state :read-only
+          :depth 1)
+         (vector
+          :translation vector :codes (,sb-vm:complex-vector-widetag)
+          :direct-superclasses (array sequence)
+          :inherits (array sequence)
+          :prototype-form (make-array 0 :adjustable t))
+         (simple-vector
+          :translation simple-vector :codes (,sb-vm:simple-vector-widetag)
+          :direct-superclasses (vector simple-array)
+          :inherits (vector simple-array array sequence)
+          :prototype-form (make-array 0))
+         (bit-vector
+          :translation bit-vector :codes (,sb-vm:complex-bit-vector-widetag)
+          :inherits (vector array sequence)
+          :prototype-form (make-array 0 :element-type 'bit :fill-pointer t))
+         (simple-bit-vector
+          :translation simple-bit-vector :codes (,sb-vm:simple-bit-vector-widetag)
+          :direct-superclasses (bit-vector simple-array)
+          :inherits (bit-vector vector simple-array
+                                array sequence)
+          :prototype-form #*)
+         (string
+          :translation string
+          :direct-superclasses (vector)
+          :inherits (vector array sequence)
+          :prototype-form "")
+         (simple-string
+          :translation simple-string
+          :direct-superclasses (string simple-array)
+          :inherits (string vector simple-array array sequence)
+          :prototype-form "")
+         (vector-nil
+          :translation (vector nil)
+          :inherits (vector array sequence)
+          :prototype-form (make-array 0 :element-type 'nil :fill-pointer t))
+         ;; This name is imperfect. It should be SIMPLE-RANK1-ARRAY-NIL
+         ;; to clearly convey that the dimensions are '(*) and not '*.
+         (simple-array-nil
+          :translation (simple-array nil (*))
+          :codes (,sb-vm:simple-array-nil-widetag)
+          :direct-superclasses (vector-nil)
+          :inherits (vector-nil vector simple-array array sequence)
+          :prototype-form (make-array 0 :element-type 'nil))
+         (base-string
+          :translation base-string
+          :codes (,sb-vm:complex-base-string-widetag)
+          :direct-superclasses (string)
+          :inherits (string vector array sequence)
+          :prototype-form (make-array 0 :element-type 'base-char :fill-pointer t))
+         (simple-base-string
+          :translation simple-base-string
+          :codes (,sb-vm:simple-base-string-widetag)
+          :direct-superclasses (base-string simple-string)
+          :inherits (base-string simple-string string vector simple-array
+                                 array sequence)
+          :prototype-form (make-array 0 :element-type 'base-char))
+         #+sb-unicode
+         (character-string
+          :translation (vector character)
+          :codes (,sb-vm:complex-character-string-widetag)
+          :direct-superclasses (string)
+          :inherits (string vector array sequence)
+          :prototype-form (make-array 0 :element-type 'character :fill-pointer t))
+         #+sb-unicode
+         (simple-character-string
+          :translation (simple-array character (*))
+          :codes (,sb-vm:simple-character-string-widetag)
+          :direct-superclasses (character-string simple-string)
+          :inherits (character-string simple-string string vector simple-array
+                                      array sequence)
+          :prototype-form (make-array 0 :element-type 'character))
+         (list
+          :translation (or cons (member nil))
+          :inherits (sequence)
+          :prototype-form 'nil)
+         (cons
+          :codes (,sb-vm:list-pointer-lowtag)
+          :translation cons
+          :inherits (list sequence)
+          :prototype-form (cons nil nil))
+         (null
+          :translation (member nil)
+          :inherits (symbol list sequence)
+          :direct-superclasses (symbol list)
+          :prototype-form 'nil)
 
-     (sb-pcl::slot-object
-      :translation (or structure-object standard-object condition)
-      :predicate slot-object-p
-      :hierarchical-p nil
-      :state :read-only
-      :prototype-form (make-defstruct-description t 'arbitrary))
+         (sb-pcl::slot-object
+          :translation (or structure-object standard-object condition)
+          :predicate slot-object-p
+          :hierarchical-p nil
+          :state :read-only
+          ;; Why is this its prototype-form? It sure looks like SLOT-OBJECT's
+          ;; metaobject has a valid prototype object that is the canonical
+          ;; CLOS object (the usual 2-word thing of layout + slot vector)
+          :prototype-form (make-defstruct-description 'arbitrary 0))
 
-     ;; KLUDGE: the length must match the subsequent defstruct.
-     (pathname :depth 1
-               :predicate pathnamep
-               :length ,(+ 7 sb-vm:instance-data-start)
-               :prototype-form (make-trivial-default-pathname))
-     (logical-pathname :depth 2
-                       :predicate logical-pathname-p
-                       :length ,(+ 7 sb-vm:instance-data-start)
-                       :prototype-form (make-trivial-default-logical-pathname)
-                       :inherits (pathname))
+         ;; KLUDGE: the length must match the subsequent defstruct.
+         (pathname :depth 1
+                   :predicate pathnamep
+                   :length ,(+ 7 sb-vm:instance-data-start)
+                   :prototype-form (make-trivial-default-pathname))
+         (logical-pathname :depth 2
+                           :predicate logical-pathname-p
+                           :length ,(+ 7 sb-vm:instance-data-start)
+                           :prototype-form (make-trivial-default-logical-pathname)
+                           :inherits (pathname))
 
-     ;; These last few are strange. STREAM has only T as an ancestor,
-     ;; so you'd think it would be at depth 1. FILE- and STRING-STREAM
-     ;; each have STREAM and T as ancestors, so you'd think they'd be at depth
-     ;; 1 greater than STREAM, instead of 2 greater. But changing any of
-     ;; these to the "obvious" value makes various type checks go wrong.
-     ;;
-     ;; Essentially the hardwiring corresponds to the indices of the
-     ;; respective types in the inherits vector for FD-STREAM.
-     ;;  * (wrapper-inherits (find-layout 'fd-stream))
-     ;;  #(#<LAYOUT for T {50300003}>
-     ;;    #<LAYOUT for STRUCTURE-OBJECT {50300103}>
-     ;;    #<LAYOUT for STREAM {50301003}>
-     ;;    #<LAYOUT for ANSI-STREAM {50301183}>
-     ;;    #<LAYOUT for FILE-STREAM {50303303}>)
+         ;; These last few are strange. STREAM has only T as an ancestor,
+         ;; so you'd think it would be at depth 1. FILE- and STRING-STREAM
+         ;; each have STREAM and T as ancestors, so you'd think they'd be at depth
+         ;; 1 greater than STREAM, instead of 2 greater. But changing any of
+         ;; these to the "obvious" value makes various type checks go wrong.
+         ;;
+         ;; Essentially the hardwiring corresponds to the indices of the
+         ;; respective types in the inherits vector for FD-STREAM.
+         ;;  * (wrapper-inherits (find-layout 'fd-stream))
+         ;;  #(#<LAYOUT for T {50300003}>
+         ;;    #<LAYOUT for STRUCTURE-OBJECT {50300103}>
+         ;;    #<LAYOUT for STREAM {50301003}>
+         ;;    #<LAYOUT for ANSI-STREAM {50301183}>
+         ;;    #<LAYOUT for FILE-STREAM {50303303}>)
 
-     (stream
-      :predicate streamp
-      :state :read-only
-      :depth 2)
-     (file-stream
-      :predicate file-stream-p
-      :state :read-only
-      :depth 4
-      :inherits (stream))
-     (string-stream
-      :predicate string-stream-p
-      :state :read-only
-      :depth 4
-      :inherits (stream))
-     ,@(loop for x across sb-vm:*specialized-array-element-type-properties*
-             unless (member (sb-vm:saetp-specifier x) '(t character base-char nil bit))
-             collect
-             ;; I'm not sure if it's an accident that there are distinct SB-KERNEL
-             ;; versus SB-VM symbols for the specialized arrays. The former are types
-             ;; in the language, and the latter are primitive object types,
-             ;; but istm they should be designated by the same symbols.
-             `(,(intern (string (sb-vm:saetp-primitive-type-name x)) *package*)
-               :translation (simple-array ,(sb-vm:saetp-specifier x) (*))
-               :codes (,(sb-vm:saetp-typecode x))
-               :direct-superclasses (vector simple-array)
-               :inherits (vector simple-array array sequence)
-               :prototype-form
-               (logically-readonlyize
-                (make-array 0 :element-type ',(sb-vm:saetp-specifier x)))))))
+         (stream
+          :predicate streamp
+          :state :read-only
+          :depth 2)
+         (file-stream
+          :predicate file-stream-p
+          :state :read-only
+          :depth 4
+          :inherits (stream))
+         (string-stream
+          :predicate string-stream-p
+          :state :read-only
+          :depth 4
+          :inherits (stream))
+         ,@(loop for x across sb-vm:*specialized-array-element-type-properties*
+                 unless (member (sb-vm:saetp-specifier x) '(t character base-char nil bit))
+                   collect
+                 ;; I'm not sure if it's an accident that there are distinct SB-KERNEL
+                 ;; versus SB-VM symbols for the specialized arrays. The former are types
+                 ;; in the language, and the latter are primitive object types,
+                 ;; but istm they should be designated by the same symbols.
+                 `(,(intern (string (sb-vm:saetp-primitive-type-name x)) *package*)
+                   :translation (simple-array ,(sb-vm:saetp-specifier x) (*))
+                   :codes (,(sb-vm:saetp-typecode x))
+                   :direct-superclasses (vector simple-array)
+                   :inherits (vector simple-array array sequence)
+                   :prototype-form
+                   (logically-readonlyize
+                    (make-array 0 :element-type ',(sb-vm:saetp-specifier x))))))))))
 
 (eval-when (#-sb-xc-host :compile-toplevel)
   (defun compute-builtin-classoids ()
@@ -1261,10 +1302,10 @@ between the ~A definition and the ~A definition"
                      (classoid (find-classoid name))
                      (translation (built-in-classoid-translation classoid))
                      (predicate
-                      (if (member name '(t random-class))
-                          'error
-                          (or (getf (cdr x) :predicate)
-                              (sb-c::backend-type-predicate translation)))))
+                       (if (member name '(t random-class))
+                           'error
+                           (or (getf (cdr x) :predicate)
+                               (sb-c::backend-type-predicate translation)))))
                 (assert predicate)
                 ;; destructuring-bind will see the first :translation
                 ;; keyword; we don't need to delete the other one.
