@@ -136,7 +136,8 @@
 
   (/show0 "entering !COLD-INIT")
   #+sb-show (setq */show* t)
-  (setq sb-vm::*codeblob-tree* nil)
+  (setq sb-vm::*immobile-codeblob-tree* nil
+        sb-vm::*dynspace-codeblob-tree* nil)
   (setq sb-kernel::*defstruct-hooks* '(sb-kernel::!bootstrap-defstruct-hook)
         sb-kernel::*struct-accesss-fragments-delayed* nil)
   (let ((stream (!make-cold-stderr-stream)))
@@ -172,10 +173,16 @@
   ;; And now *CURRENT-THREAD*
   (sb-thread::init-main-thread)
 
+  ;; not sure why this is needed on some architectures. Dark magic.
+  (setf (fdefn-fun (find-or-create-fdefn '%coerce-callable-for-call))
+        #'%coerce-callable-to-fun)
+  (show-and-call !loader-cold-init)
   ;; Assert that FBOUNDP doesn't choke when its answer is NIL.
   ;; It was fine if T because in that case the legality of the arg is certain.
   ;; And be extra paranoid - ensure that it really gets called.
-  (locally (declare (notinline fboundp)) (fboundp '(setf !zzzzzz)))
+  (locally
+      (declare (notinline fboundp) (optimize safety)) ; is unsafely flushable
+    (fboundp '(setf !zzzzzz)))
 
   ;; Printing of symbols requires that packages be filled in, because
   ;; OUTPUT-SYMBOL calls FIND-SYMBOL to determine accessibility. Also
@@ -247,6 +254,12 @@
 
   (makunbound '*!cold-toplevels*) ; so it gets GC'd
 
+  ;; Need the static-space replica of the assembly routine jump vector
+  ;; filled in, and the static space vector of static fdefns.
+  ;; This matters only for code that gets compiled to dynamic space,
+  ;; so it's OK that it occurs somewhat late in cold-init.
+  #+x86-64 (sb-vm::validate-asm-routine-vector)
+
   #+win32 (show-and-call reinit-internal-real-time)
 
   ;; Set sane values again, so that the user sees sane values instead
@@ -267,7 +280,6 @@
   (mapcar #'proclaim sb-c::*queued-proclaims*)
   (makunbound 'sb-c::*queued-proclaims*)
 
-  (show-and-call !loader-cold-init)
   (show-and-call os-cold-init-or-reinit)
   (show-and-call !pathname-cold-init)
 
@@ -390,7 +402,9 @@ process to continue normally."
     ;; can be called, as pretty much anything can assume that it is set.
     (when total ; newly started process, and not a failed save attempt
       (sb-thread::init-main-thread)
-      (rebuild-package-vector))
+      #+x86-64 (sb-vm::validate-asm-routine-vector)
+      (rebuild-package-vector)
+      (rebuild-pathname-cache))
     ;; Initializing the standard streams calls ALLOC-BUFFER which calls FINALIZE
     (finalizers-reinit)
     ;; Initialize streams next, so that any errors can be printed
@@ -408,7 +422,8 @@ process to continue normally."
   (when (eq *invoke-debugger-hook* 'sb-debug::debugger-disabled-hook)
     (sb-debug::disable-debugger))
   (call-hooks "initialization" *init-hooks*)
-  #+sb-thread (finalizer-thread-start))
+  #+sb-thread (finalizer-thread-start)
+  (sb-vm::!setup-cpu-specific-routines))
 
 ;;;; some support for any hapless wretches who end up debugging cold
 ;;;; init code
