@@ -25,7 +25,7 @@
 (defmacro finalizer-max-id (store) `(elt ,store 2))
 
 (defun make-finalizer-store (array-length)
-  (let* ((v (make-array (the index array-length)))
+  (let* ((v (make-array (the index array-length) :initial-element 0))
          (ht (make-system-hash-table :test 'eq :weakness :key :synchronized nil
                                      :finalizer t)))
     ;; The recycle bin has a dummy item in front so that the simple-vector
@@ -43,8 +43,7 @@
 (declaim (simple-vector **finalizer-store**))
 
 (defun finalize (object function &key dont-save
-                        &aux (function (%coerce-callable-to-fun function))
-                             (item (if dont-save (list function) function)))
+                        &aux (function (%coerce-callable-to-fun function)))
   "Arrange for the designated FUNCTION to be called when there
 are no more references to OBJECT, including references in
 FUNCTION itself.
@@ -91,11 +90,14 @@ Examples:
     (finalize \"oops\" #'oops)
     (oops)) ; GC causes re-entry to #'oops due to the finalizer
             ; -> ERROR, caught, WARNING signalled"
+  (declare (sb-c::tlab :system))
   (unless object
     (error "Cannot finalize NIL."))
-  (with-finalizer-store (store)
-    (let ((id (gethash object (finalizer-id-map store))))
-      (cond (id ; object already has at least one finalizer
+  (let ((item (if dont-save (list function) function)))
+    (with-finalizer-store (store)
+      (let ((id (gethash object (finalizer-id-map store))))
+        (cond
+            (id ; object already has at least one finalizer
              ;; Multiple finalizers are invoked in the order added.
              (let* ((old (svref store id))
                     (new (make-array (if (simple-vector-p old)
@@ -130,7 +132,7 @@ Examples:
              ;; Clear out lingering junk from (SVREF STORE ID) before
              ;; establishing that OBJECT maps to that index.
              (setf (svref store id) item
-                   (gethash object (finalizer-id-map store)) id)))))
+                   (gethash object (finalizer-id-map store)) id))))))
   object)
 
 (defun invalidate-fd-streams ()
@@ -230,8 +232,9 @@ Examples:
          ;; Not strictly necessary to do this: the next FINALIZE claiming
          ;; the same ID would assign a fresh list anyway.
          (setf (svref store it) 0)
-         (atomic-push it (finalizer-recycle-bin store)))))
-    object))
+         (locally (declare (sb-c::tlab :system))
+           (atomic-push it (finalizer-recycle-bin store)))))))
+  object)
 
 ;;; Drain the queue of finalizers and return when empty.
 ;;; Concurrent invocations of this function in different threads are ok.
@@ -239,21 +242,29 @@ Examples:
 ;;; See the trace at the bottom of this file.
 (defvar *in-a-finalizer* nil)
 #+sb-thread (define-alien-variable finalizer-thread-runflag int)
-(defun run-pending-finalizers ()
+
+(defun run-pending-finalizers (&aux (hashtable (finalizer-id-map **finalizer-store**))
+                                    (ran-a-system-finalizer)
+                                    (system-finalizer-scratchpad (list 0))
+                                    (ran-a-user-finalizer))
+  (declare (truly-dynamic-extent system-finalizer-scratchpad))
   ;; This never acquires the finalizer store lock. Code accordingly.
-  (let ((hashtable (finalizer-id-map **finalizer-store**)))
-    (loop
-     ;; Perform no further work if trying to stop the thread, even if there is work.
-     #+sb-thread (when (zerop finalizer-thread-runflag) (return))
-     (let ((cell (hash-table-culled-values hashtable)))
-       ;; This is like atomic-pop, but its obtains the first cons cell
-       ;; in the list, not the car of the first cons.
-       ;; Possible TODO: when no other work remains, free the *JOINABLE-THREADS*,
-       ;; though MAKE-THREAD and JOIN-THREAD do that also, so there's no memory leak.
-       (loop (unless cell (return-from run-pending-finalizers))
-             (let ((actual (cas (hash-table-culled-values hashtable)
-                                cell (cdr cell))))
-               (if (eq actual cell) (return) (setq cell actual))))
+  (loop
+   ;; Perform no further work if trying to stop the thread, even if there is work.
+   #+sb-thread (when (zerop finalizer-thread-runflag) (return))
+   ;; Try to run 1 system finalizer
+   (setq ran-a-system-finalizer (sb-vm::immobile-code-dealloc-1 system-finalizer-scratchpad))
+   ;; Try to run 1 user finalizer
+   (let ((cell (hash-table-culled-values hashtable)))
+     ;; This is like atomic-pop, but its obtains the first cons cell
+     ;; in the list, not the car of the first cons.
+     ;; Possible TODO: when no other work remains, free the *JOINABLE-THREADS*,
+     ;; though MAKE-THREAD and JOIN-THREAD do that also, so there's no memory leak.
+     (loop (unless cell (return))
+           (let ((actual (cas (hash-table-culled-values hashtable)
+                              cell (cdr cell))))
+             (if (eq actual cell) (return) (setq cell actual))))
+     (when cell
        (let* ((id (the index (car cell)))
               ;; No other thread can modify **FINALIZER-STORE** at index ID
               ;; because the table no longer contains an object mapping to
@@ -280,6 +291,7 @@ Examples:
            (if (simple-vector-p finalizers)
                (map nil #'call finalizers)
                (call finalizers)))
+         (setq ran-a-user-finalizer t)
          ;; While the assignment to (SVREF STORE ID) should have been adequate,
          ;; we don't know that the vector is current - a new vector could have
          ;; gotten assigned into **FINALIZER-STORE** in between [1] and [2],
@@ -297,7 +309,11 @@ Examples:
          ;; removing dangling references, but if it's just a function,
          ;; there's nothing to smash.
          (cond ((simple-vector-p finalizers) (fill finalizers 0))
-               ((consp finalizers) (rplaca finalizers 0))))))))
+               ((consp finalizers) (rplaca finalizers 0))))))
+   ;; Did this iteration do anything at all?
+   (unless (or ran-a-system-finalizer ran-a-user-finalizer) (return))
+   (setq ran-a-system-finalizer nil
+         ran-a-user-finalizer nil)))
 
 (define-load-time-global *finalizer-thread* nil)
 (declaim (type (or sb-thread:thread (eql :start) null) *finalizer-thread*))
