@@ -31,6 +31,7 @@
 (define-alien-routine "gc_and_save" void
   (file c-string)
   (prepend-runtime int)
+  (purify int)
   (save-runtime-options int)
   (compressed int)
   (compression-level int)
@@ -81,15 +82,15 @@
        sb-kernel::*gc-epoch*))
 
 (defun start-lisp (toplevel callable-exports)
-  (named-lambda %start-lisp ()
-    (cond (callable-exports
-           (reinit t)
-           (dolist (export callable-exports)
-             (sb-alien::initialize-alien-callable-symbol export)))
-          (t
-           (handling-end-of-the-world
-            (reinit t)
-            (funcall toplevel))))))
+  (if callable-exports
+      (named-lambda %start-lisp ()
+        (reinit t)
+        (dolist (export callable-exports)
+          (sb-alien::initialize-alien-callable-symbol export)))
+      (named-lambda %start-lisp ()
+        (handling-end-of-the-world
+          (reinit t)
+          (funcall toplevel)))))
 
 (defun save-lisp-and-die (core-file-name &key
                                          (toplevel #'toplevel-init toplevel-supplied)
@@ -139,13 +140,10 @@ The following &KEY arguments are defined:
      :TOPLEVEL argument cannot be supplied.
 
   :PURIFY
-     If true (the default on cheneygc), do a purifying GC which moves all
-     dynamically allocated objects into static space. This takes
-     somewhat longer than the normal GC which is otherwise done, but
-     it's only done once, and subsequent GC's will be done less often
-     and will take less time in the resulting core file. See the PURIFY
-     function. This parameter has no effect on platforms using the
-     generational garbage collector.
+     If true (the default), then some objects in the restarted core will
+     be memory-mapped as read-only. Among those objects are numeric vectors
+     that were determined to be compile-time constants, and any immutable
+     values according to the language specification such as symbol names.
 
   :ROOT-STRUCTURES
      This should be a list of the main entry points in any newly loaded
@@ -166,8 +164,8 @@ The following &KEY arguments are defined:
      This is only meaningful if the runtime was built with the :SB-CORE-COMPRESSION
      feature enabled. If NIL (the default), saves to uncompressed core files. If
      :SB-CORE-COMPRESSION was enabled at build-time, the argument may also be
-     an integer from -1 to 9, corresponding to zlib compression levels, or T
-     (which is equivalent to the default compression level, -1).
+     an integer from -7 to 22, corresponding to zstd compression levels, or T
+     (which is equivalent to the default compression level, 9).
 
   :APPLICATION-TYPE
      Present only on Windows and is meaningful only with :EXECUTABLE T.
@@ -206,7 +204,7 @@ seem to be good quick fixes for either limitation and no one has been
 sufficiently motivated to do lengthy fixes."
   (declare (ignore environment-name))
   #+gencgc
-  (declare (ignore purify) (ignorable root-structures))
+  (declare (ignorable root-structures))
   (when (and callable-exports toplevel-supplied)
     (error ":TOPLEVEL cannot be supplied when there are callable exports."))
   ;; If the toplevel function is not defined, this will signal an
@@ -214,10 +212,10 @@ sufficiently motivated to do lengthy fixes."
   (let ((toplevel (%coerce-callable-to-fun toplevel))
         *streams-closed-by-slad*)
     #+sb-core-compression
-    (check-type compression (or boolean (integer -1 9)))
+    (check-type compression (or boolean (integer -7 22)))
     #-sb-core-compression
     (when compression
-      (error "Unable to save compressed core: this runtime was not built with zlib support"))
+      (error "Unable to save compressed core: this runtime was not built with zstd support"))
     (when *dribble-stream*
       (restart-case (error "Dribbling to ~s is enabled." (pathname *dribble-stream*))
         (continue ()
@@ -227,17 +225,13 @@ sufficiently motivated to do lengthy fixes."
           :report "Abort saving the core."
           (return-from save-lisp-and-die))))
     (when (eql t compression)
-      (setf compression -1))
+      (setf compression 9))
     (flet ((foreign-bool (value)
              (if value 1 0)))
       (let ((name (native-namestring (physicalize-pathname core-file-name)
                                      :as-file t))
             (startfun (start-lisp toplevel callable-exports)))
         (deinit)
-        ;; FIXME: Would it be possible to unmix the PURIFY logic from this
-        ;; function, and just do a GC :FULL T here? (Then if the user wanted
-        ;; a PURIFYed image, he'd just run PURIFY immediately before calling
-        ;; SAVE-LISP-AND-DIE.)
         #+gencgc
         (progn
           ;; Scan roots as close as possible to GC-AND-SAVE, in case anything
@@ -260,6 +254,7 @@ sufficiently motivated to do lengthy fixes."
           ;; since the GC will invalidate the stack.
           (gc-and-save name
                        (foreign-bool executable)
+                       (foreign-bool purify)
                        (foreign-bool save-runtime-options)
                        (foreign-bool compression)
                        (or compression 0)
@@ -318,7 +313,7 @@ sufficiently motivated to do lengthy fixes."
   (let (error)
     (with-system-mutex (sb-thread::*make-thread-lock*)
       (finalizer-thread-stop)
-      #+pauseless-threadstart (sb-thread::join-pthread-joinables #'identity)
+      (sb-thread::%dispose-thread-structs)
       (let ((threads (sb-thread:list-all-threads))
             (starting
              (setq sb-thread::*starting-threads* ; ordinarily pruned in MAKE-THREAD
@@ -338,10 +333,10 @@ sufficiently motivated to do lengthy fixes."
   (float-deinit)
   (profile-deinit)
   (foreign-deinit)
-  ;; To have any hope of making pathname interning actually work,
-  ;; this CLRHASH would need to be removed. But removing it causes excess
-  ;; garbage retention because weakness doesn't work. It's a catch-22.
-  (clrhash *pathnames*)
+  (when (zerop (hash-table-count sb-kernel::*forward-referenced-wrappers*))
+    ;; I think this table should always be empty but I'm not sure. If it is,
+    ;; recreate it so that we don't preserve an empty vector taking up 16KB
+    (setq sb-kernel::*forward-referenced-wrappers* (make-hash-table :test 'equal)))
   ;; Clean up the simulated weak list of covered code components.
   (rplacd sb-c:*code-coverage-info*
           (delete-if-not #'weak-pointer-value (cdr sb-c:*code-coverage-info*)))
@@ -355,8 +350,13 @@ sufficiently motivated to do lengthy fixes."
   #+immobile-code (sb-vm::statically-link-core)
   (invalidate-fd-streams)
   (finalizers-deinit)
+  ;; Try to shrink the pathname cache. It might be largely nulls
+  (rebuild-pathname-cache)
+  (sb-vm::restore-cpu-specific-routines)
   ;; Do this last, to have some hope of printing if we need to.
   (stream-deinit)
+  (setf sb-c::*compile-elapsed-time* 0
+        sb-c::*compile-file-elapsed-time* 0)
   (setf * nil ** nil *** nil
         - nil + nil ++ nil +++ nil
         /// nil // nil / nil))
@@ -476,8 +476,9 @@ sufficiently motivated to do lengthy fixes."
             (sb-vm:do-referenced-object (obj examine)
               (simple-vector
                :extend
-               (when (and written (logtest sb-vm:vector-addr-hashing-flag
-                                           (get-header-data obj)))
+               (when (and written (test-header-data-bit obj
+                                                        (ash sb-vm:vector-addr-hashing-flag
+                                                             sb-vm:array-flags-data-position)))
                  (setf (svref obj 1) 1)))))))))) ; set need-to-rehash
 
 sb-c::
@@ -633,7 +634,7 @@ sb-c::
             (callees (cdr item)))
         (dolist (callee callees)
           (push caller (gethash callee reverse-graph)))))
-    ;; Compute popularity of each code component in varyobj space
+    ;; Compute popularity of each code component in text space
     (map-allocated-objects
      (lambda (obj type size)
        (declare (ignore size))
