@@ -692,11 +692,10 @@
 ;;;
 ;;; TODO: As STRING is an especially important union type,
 ;;; it could be interned by canonicalizing its subparts into
-;;; ARRAY of {CHARACTER,BASE-CHAR,NIL} in that exact order always.
+;;; ARRAY of {CHARACTER,BASE-CHAR} in that exact order always.
 ;;; It will therefore admit quick TYPE=, but not quick failure, since
 ;;;   (type= (specifier-type '(or (simple-array (member #\a) (*))
-;;;                               (simple-array character (*))
-;;;                               (simple-array nil (*))))
+;;;                               (simple-array character (*))))
 ;;;          (specifier-type 'simple-string)) => T and T
 ;;; even though (MEMBER #\A) is not TYPE= to BASE-CHAR.
 ;;;
@@ -870,7 +869,7 @@
             (:constructor %make-simd-pack-type (element-type))
             (:copier nil))
   (element-type (missing-arg)
-   :type (cons #||(member #.*simd-pack-element-types*) ||#)
+   :type (simple-bit-vector #.(length *simd-pack-element-types*))
    :read-only t))
 
 #+sb-simd-pack-256
@@ -879,7 +878,7 @@
             (:constructor %make-simd-pack-256-type (element-type))
             (:copier nil))
   (element-type (missing-arg)
-   :type (cons #||(member #.*simd-pack-element-types*) ||#)
+   :type (simple-bit-vector #.(length *simd-pack-element-types*))
    :read-only t))
 
 (declaim (ftype (sfunction (ctype ctype) (values t t)) csubtypep))
