@@ -9,8 +9,6 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-#+riscv (setf (extern-alien "gencgc_verbose" char) 3)
-
 (in-package "COMMON-LISP-USER")
 
 ;;;; general warm init compilation policy
@@ -82,7 +80,17 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
       ;; in case there is more than one set of floating-point formats.
       (assert (eq (read stream) :default))
       (sb-kernel::with-float-traps-masked (:overflow :divide-by-zero)
-        (let ((*package* (find-package "SB-KERNEL")))
+        (let ((*readtable* (copy-readtable))
+              (*package* (find-package "SB-KERNEL")))
+          (set-dispatch-macro-character
+           #\# #\. (lambda (stream subchar arg)
+                     (declare (ignore subchar arg))
+                     (let ((expr (read stream t nil t)))
+                       (ecase (car expr)
+                         (sb-kernel:make-single-float
+                          (sb-kernel:make-single-float (second expr)))
+                         (sb-kernel:make-double-float
+                          (sb-kernel:make-double-float (second expr) (third expr)))))))
           (dolist (expr (read stream))
             (destructuring-bind (fun args . result) expr
               (let ((result (if (eq (first result) 'sb-kernel::&values)
@@ -162,6 +170,7 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                     (ecase (if (boundp '*compile-files-p*) *compile-files-p* t)
                      ((t)
                       (let ((sb-c::*source-namestring* fullname)
+                            (sb-c::*force-system-tlab* (search "src/pcl" stem))
                             (sb-ext:*derive-function-types*
                               (unless (search "/pcl/" stem)
                                 t)))
@@ -234,7 +243,8 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
     (let (list)
       (sb-int:dohash ((name vop) sb-c::*backend-parsed-vops*)
         (declare (ignore vop))
-        (push (cons (gethash name sb-c::*static-vop-usage-counts* 0) name) list))
+        (unless (char= (char (string name) 0) #\!)
+          (push (cons (gethash name sb-c::*static-vop-usage-counts* 0) name) list)))
       (dolist (cell (sort list #'> :key #'car))
         (format output "~7d ~s~%" (car cell) (cdr cell))))))
 
