@@ -205,6 +205,10 @@
         (setf (conset-max conset) (1+ number))))
     conset)
 
+  (defun conset-delete (constraint conset)
+    (setf (sbit (conset-vector conset) (%constraint-number constraint)) 0)
+    conset)
+
   (defun conset= (conset1 conset2)
     (let* ((vector1 (conset-vector conset1))
            (vector2 (conset-vector conset2))
@@ -320,7 +324,6 @@
 ;;; through the intersection of such constraints and the constraints in a conset.
 ;;;
 ;;; EQL-var constraints assert that two lambda-vars are EQL.
-;;; Private constraints assert that a lambda-var is EQL or not EQL to a constant.
 ;;; Inheritable constraints are constraints that may be propagated to EQL
 ;;; lambda-vars (along with EQL-var constraints).
 ;;;
@@ -357,7 +360,7 @@
              (setf (car cons) con)))
        (typecase y
          (constant
-          (let ((vec (ensure-vec (lambda-var-private-constraints x))))
+          (let ((vec (ensure-vec (lambda-var-inheritable-constraints x))))
             (vector-push-extend con vec)))
          (lambda-var
           (let ((vec (if (or (constraint-not-p con)
@@ -485,9 +488,6 @@
     `(block nil
        (flet ((body-fun (,symbol)
                 ,@body))
-         (do-conset-constraints-intersection
-             (con (,conset (lambda-var-private-constraints ,variable)))
-           (body-fun con))
          (do-conset-constraints-intersection
              (con (,conset (lambda-var-eql-var-constraints ,variable)))
            (body-fun con))
@@ -704,105 +704,110 @@
     (let ((consequent-constraints (make-conset))
           (alternative-constraints (make-conset))
           (quick-p (policy if (> compilation-speed speed))))
-      (macrolet ((add (fun x y not-p)
-                   `(add-complement-constraints quick-p
-                                                ,fun ,x ,y ,not-p
-                                                constraints
-                                                consequent-constraints
-                                                alternative-constraints)))
-        (typecase use
-          (ref
-           (add 'typep (ok-lvar-lambda-var (ref-lvar use) constraints)
-                (specifier-type 'null) t))
-          (combination
-           (unless (eq (combination-kind use)
-                       :error)
-             (let ((name (uncross
-                          (lvar-fun-name
-                           (basic-combination-fun use))))
-                   (args (basic-combination-args use)))
-               (add-equality-constraints name args
-                                         constraints consequent-constraints alternative-constraints)
-               (case name
-                 ((%typep %instance-typep)
-                  (let ((type (second args)))
-                    (when (constant-lvar-p type)
-                      (let ((val (lvar-value type)))
-                        (add 'typep
-                             (ok-lvar-lambda-var (first args) constraints)
-                             (if (ctype-p val)
-                                 val
-                                 (let ((*compiler-error-context* use))
-                                   (specifier-type val)))
-                             nil)))))
-                 ((eq eql)
-                  (let* ((arg1 (first args))
-                         (var1 (ok-lvar-lambda-var arg1 constraints))
-                         (arg2 (second args))
-                         (var2 (ok-lvar-lambda-var arg2 constraints)))
-                    ;; The code below assumes that the constant is the
-                    ;; second argument in case of variable to constant
-                    ;; comparison which is sometimes true (see source
-                    ;; transformations for EQ, EQL and CHAR=). Fixing
-                    ;; that would result in more constant substitutions
-                    ;; which is not a universally good thing, thus the
-                    ;; unnatural asymmetry of the tests.
-                    (cond ((not var1)
-                           (when var2
-                             (add-test-constraint quick-p
-                                                  'typep var2 (lvar-type arg1)
-                                                  nil constraints
-                                                  consequent-constraints)))
-                          (var2
-                           (add 'eql var1 var2 nil))
-                          ((constant-lvar-p arg2)
-                           (add 'eql var1
-                                (nth-value 1 (lvar-value arg2))
-                                nil))
-                          (t
-                           (add-test-constraint quick-p
-                                                'typep var1 (lvar-type arg2)
-                                                nil constraints
-                                                consequent-constraints)))))
-                 ((< >)
-                  (when (= (length args) 2)
-                    (flet ((handle-array-in-bounds-p (index-arg index-var length-arg)
-                             (multiple-value-bind (kind x y)
-                                 (array-in-bounds-p-constraints constraints index-arg index-var
-                                                                length-arg)
-                               (when kind
-                                 (add-test-constraint quick-p
-                                                      kind x y
-                                                      nil constraints
-                                                      consequent-constraints)))))
-                      (let* ((arg1 (first args))
-                             (var1 (ok-lvar-lambda-var arg1 constraints))
-                             (arg2 (second args))
-                             (var2 (ok-lvar-lambda-var arg2 constraints)))
+      (labels ((add (fun x y not-p)
+                 (add-complement-constraints quick-p
+                                             fun x y not-p
+                                             constraints
+                                             consequent-constraints
+                                             alternative-constraints))
+               (process-node (node)
+                 (typecase node
+                   (ref
+                    (add 'typep (ok-lvar-lambda-var (ref-lvar node) constraints)
+                         (specifier-type 'null) t)
+                    (let ((use (principal-lvar-ref-use (ref-lvar use))))
+                      (when (and use
+                                 (not (ref-p use)))
+                        (process-node use))))
+                   (combination
+                    (unless (eq (combination-kind node) :error)
+                      (let ((name (uncross
+                                   (lvar-fun-name
+                                    (basic-combination-fun node))))
+                            (args (basic-combination-args node)))
+                        (add-equality-constraints name args
+                                                  constraints consequent-constraints alternative-constraints)
                         (case name
-                          (<
-                           (handle-array-in-bounds-p arg1 var1 arg2))
-                          (>
-                           (handle-array-in-bounds-p arg2 var2 arg1)))
-                        (when var1
-                          (add name var1 (lvar-type arg2) nil))
-                        (when var2
-                          (add (if (eq name '<) '> '<) var2 (lvar-type arg1) nil))))))
-                 (=
-                  (when (= (length args) 2)
-                    (let* ((arg1 (first args))
-                           (var1 (ok-lvar-lambda-var arg1 constraints))
-                           (arg2 (second args))
-                           (var2 (ok-lvar-lambda-var arg2 constraints)))
-                      (when var1
-                        (add name var1 (lvar-type arg2) nil))
-                      (when var2
-                        (add name var2 (lvar-type arg1) nil)))))
-                 (t
-                  (add-combination-test-constraints use constraints
-                                                    consequent-constraints
-                                                    alternative-constraints
-                                                    quick-p))))))))
+                          ((%typep %instance-typep)
+                           (let ((type (second args)))
+                             (when (constant-lvar-p type)
+                               (let ((val (lvar-value type)))
+                                 (add 'typep
+                                      (ok-lvar-lambda-var (first args) constraints)
+                                      (if (ctype-p val)
+                                          val
+                                          (let ((*compiler-error-context* node))
+                                            (specifier-type val)))
+                                      nil)))))
+                          ((eq eql)
+                           (let* ((arg1 (first args))
+                                  (var1 (ok-lvar-lambda-var arg1 constraints))
+                                  (arg2 (second args))
+                                  (var2 (ok-lvar-lambda-var arg2 constraints)))
+                             ;; The code below assumes that the constant is the
+                             ;; second argument in case of variable to constant
+                             ;; comparison which is sometimes true (see source
+                             ;; transformations for EQ, EQL and CHAR=). Fixing
+                             ;; that would result in more constant substitutions
+                             ;; which is not a universally good thing, thus the
+                             ;; unnatural asymmetry of the tests.
+                             (cond ((not var1)
+                                    (when var2
+                                      (add-test-constraint quick-p
+                                                           'typep var2 (lvar-type arg1)
+                                                           nil constraints
+                                                           consequent-constraints)))
+                                   (var2
+                                    (add 'eql var1 var2 nil))
+                                   ((constant-lvar-p arg2)
+                                    (add 'eql var1
+                                         (nth-value 1 (lvar-value arg2))
+                                         nil))
+                                   (t
+                                    (add-test-constraint quick-p
+                                                         'typep var1 (lvar-type arg2)
+                                                         nil constraints
+                                                         consequent-constraints)))))
+                          ((< >)
+                           (when (= (length args) 2)
+                             (flet ((handle-array-in-bounds-p (index-arg index-var length-arg)
+                                      (multiple-value-bind (kind x y)
+                                          (array-in-bounds-p-constraints constraints index-arg index-var
+                                                                         length-arg)
+                                        (when kind
+                                          (add-test-constraint quick-p
+                                                               kind x y
+                                                               nil constraints
+                                                               consequent-constraints)))))
+                               (let* ((arg1 (first args))
+                                      (var1 (ok-lvar-lambda-var arg1 constraints))
+                                      (arg2 (second args))
+                                      (var2 (ok-lvar-lambda-var arg2 constraints)))
+                                 (case name
+                                   (<
+                                    (handle-array-in-bounds-p arg1 var1 arg2))
+                                   (>
+                                    (handle-array-in-bounds-p arg2 var2 arg1)))
+                                 (when var1
+                                   (add name var1 (lvar-type arg2) nil))
+                                 (when var2
+                                   (add (if (eq name '<) '> '<) var2 (lvar-type arg1) nil))))))
+                          (=
+                           (when (= (length args) 2)
+                             (let* ((arg1 (first args))
+                                    (var1 (ok-lvar-lambda-var arg1 constraints))
+                                    (arg2 (second args))
+                                    (var2 (ok-lvar-lambda-var arg2 constraints)))
+                               (when var1
+                                 (add name var1 (lvar-type arg2) nil))
+                               (when var2
+                                 (add name var2 (lvar-type arg1) nil)))))
+                          (t
+                           (add-combination-test-constraints node constraints
+                                                             consequent-constraints
+                                                             alternative-constraints
+                                                             quick-p)))))))))
+        (process-node use))
       (values consequent-constraints alternative-constraints))))
 
 ;;;; Applying constraints
@@ -1080,7 +1085,7 @@
                  (if (member-type-p other)
                      (mapc-member-type-members #'note-not other)
                      (setq not-res (type-union not-res other)))
-                 (setq res (type-approx-intersection2 res other))))
+                 (setq res (type-intersection res other))))
             (array-in-bounds-p
              (unless (eq (ref-constraints ref)
                          (pushnew con (ref-constraints ref)))
@@ -1109,8 +1114,7 @@
                         (change-ref-leaf ref other)
                         (when (constant-p other) (return)))
                        (t
-                        (setq res (type-approx-intersection2
-                                   res other-type))))))))
+                        (setq res (type-intersection res other-type))))))))
             ((< >)
              (let* ((greater (eq kind '>))
                     (greater (if not-p (not greater) greater)))
@@ -1125,17 +1129,17 @@
                   (let ((type (constrain-real-to-integer y greater not-p)))
                     (when type
                       (setf res
-                            (type-approx-intersection2 res type))))))))
+                            (type-intersection res type))))))))
             (=
              (when (and (numeric-type-p y)
                         (not not-p))
                (setf res
-                     (type-approx-intersection2 res
-                                                (type-union (make-numeric-type :low (numeric-type-low y)
-                                                                               :high (numeric-type-high y))
-                                                            (make-numeric-type :complexp :complex
-                                                                               :low (numeric-type-low y)
-                                                                               :high (numeric-type-high y)))))))))))
+                     (type-intersection res
+                                        (type-union (make-numeric-type :low (numeric-type-low y)
+                                                                       :high (numeric-type-high y))
+                                                    (make-numeric-type :complexp :complex
+                                                                       :low (numeric-type-low y)
+                                                                       :high (numeric-type-high y)))))))))))
     (cond ((and (if-p (node-dest ref))
                 (or (xset-member-p nil not-set)
                     (csubtypep (specifier-type 'null) not-res)))
@@ -1368,36 +1372,44 @@
 ;;; Join the constraints coming from the predecessors of BLOCK on
 ;;; every constrained variable into the constraint set IN.
 (defun join-type-constraints (in block)
-  (let ((vars '()))
-    ;; Find some set of constrained variables in the predecessors.
-    (dolist (pred (block-pred block))
-      (let ((out (block-out-for-successor pred block)))
-        (when out
-          (do-conset-elements (con out)
-            (let ((kind  (constraint-kind con))
-                  (y     (constraint-y con))
-                  (not-p (constraint-not-p con)))
-              (when (or (member kind '(typep < >))
-                        (and (eq kind 'eql) (or (not not-p)
-                                                (constant-p y)))
-                        (and (eq kind '=) (and (numeric-type-p y)
-                                               (not not-p))))
-                (pushnew (constraint-x con) vars))))
-          (return))))
-    (dolist (var vars)
-      (let ((in-var-type *empty-type*))
-        (dolist (pred (block-pred block))
-          (let ((out (block-out-for-successor pred block)))
-            (when out
-              (setq in-var-type
-                    (type-union in-var-type
-                                (type-from-constraints var out *universal-type*))))))
-        (unless (eq in-var-type *universal-type*)
-          (conset-adjoin (find-or-create-constraint 'typep
-                                                    var
-                                                    in-var-type
-                                                    nil)
-                         in))))))
+  (let ((vars '())
+        (predecessors (block-pred block)))
+    (progn
+      ;;unless (singleton-p predecessors) for some reason it makes the whole thing loop infinitely
+      ;; Find some set of constrained variables in the predecessors.
+      (dolist (pred predecessors)
+        (let ((out (block-out-for-successor pred block)))
+          (when out
+            (do-conset-elements (con out)
+              (let ((kind  (constraint-kind con))
+                    (y     (constraint-y con))
+                    (not-p (constraint-not-p con)))
+                (when (or (member kind '(typep < >))
+                          (and (eq kind 'eql) (or (not not-p)
+                                                  (constant-p y)))
+                          (and (eq kind '=) (and (numeric-type-p y)
+                                                 (not not-p))))
+                  (pushnew (constraint-x con) vars))))
+            (return))))
+      (dolist (var vars)
+        (let ((in-var-type *empty-type*))
+          (dolist (pred (block-pred block))
+            (let ((out (block-out-for-successor pred block)))
+              (when out
+                (setq in-var-type
+                      (type-union in-var-type
+                                  (type-from-constraints var out *universal-type*))))))
+
+          (unless (eq in-var-type *universal-type*)
+            ;; Remove the existing constraints to avoid joining them again later.
+            (do-propagatable-constraints (con (in var))
+              (when (eq (constraint-kind con) 'typep)
+                (conset-delete con in)))
+            (conset-adjoin (find-or-create-constraint 'typep
+                                                      var
+                                                      in-var-type
+                                                      nil)
+                           in)))))))
 
 (defun compute-block-in (block join-types-p)
   (let ((in nil))
