@@ -11,6 +11,24 @@
 
 (in-package "SB-C")
 
+;;;; DEFTYPEs
+
+;;; An INLINEP value describes how a function is called. The values
+;;; have these meanings:
+;;;     NIL     No declaration seen: do whatever you feel like, but don't
+;;;             dump an inline expansion.
+;;; NOTINLINE  NOTINLINE declaration seen: always do full function call.
+;;;    INLINE  INLINE declaration seen: save expansion, expanding to it
+;;;             if policy favors.
+;;; MAYBE-INLINE
+;;;             Retain expansion, but only use it opportunistically.
+;;;             MAYBE-INLINE is quite different from INLINE. As explained
+;;;             by APD on #lisp 2005-11-26: "MAYBE-INLINE lambda is
+;;;             instantiated once per component, INLINE - for all
+;;;             references (even under #'without FUNCALL)."
+(deftype inlinep () '(member inline maybe-inline notinline nil))
+
+
 ;;;; source-hacking defining forms
 
 ;;; Parse a DEFMACRO-style lambda-list, setting things up so that a
@@ -362,7 +380,7 @@
 ;;; methods are passed an additional POLICY argument, and IR2-CONVERT
 ;;; methods are passed an additional IR2-BLOCK argument.
 (defmacro defoptimizer (what (lambda-list
-                              &optional (node (sb-xc:gensym) node-p)
+                              &optional (node (gensym))
                               &rest vars)
                         &body body)
   (let ((name (flet ((function-name (name)
@@ -384,38 +402,55 @@
              ,@body)
            ,@(loop for vop-name in (ensure-list (second what))
                    collect
-                   `(setf (vop-info-optimizer (template-or-lose ',vop-name))
-                          #',name)))
+                   `(set-vop-optimizer (template-or-lose ',vop-name)
+                                       ,(if (cddr what)
+                                            `(cons #',name ',(caddr what))
+                                            `#',name))))
         (binding* (((forms decls) (parse-body body nil))
                    ((var-decls more-decls) (extract-var-decls decls vars))
-                   ;; In case the BODY declares IGNORE of the formal NODE var,
-                   ;; we rebind it from N-NODE and never reference it from BINDS.
-                   (n-node (make-symbol "NODE"))
-                   ((binds lambda-vars gensyms)
-                    (parse-deftransform lambda-list n-node
+                   ((binds lambda-vars)
+                    (parse-deftransform lambda-list node
                                         `(return-from ,name
                                            ,(if (and (consp what)
                                                      (eq (second what)
                                                          'equality-constraint))
                                                 :give-up
-                                                nil)))))
+                                                nil))))
+                   (args (gensym)))
           (declare (ignore lambda-vars))
           `(progn
              ;; We can't stuff the BINDS as &AUX vars into the lambda list
              ;; because there can be a RETURN-FROM in there.
-             (defun ,name (,n-node ,@vars)
+             (defun ,name (,node ,@vars &rest ,args)
+               (declare (ignorable ,node ,@(butlast vars))
+                        (ignore ,args))
                ,@(if var-decls (list var-decls))
-               (let* (,@binds ,@(if node-p `((,node ,n-node))))
-                 ;; Syntax requires naming NODE even if undesired if VARS
-                 ;; are present, so in that case make NODE ignorable.
-                 (declare (ignorable ,@(if (and vars node-p) `(,node))
-                                     ,@gensyms))
+               (let* (,@binds)
+                 (declare (ignorable ,@(mapcar #'car binds)))
                  ,@more-decls ,@forms))
              ,@(when (consp what)
                  `((setf (,(let ((*package* (sb-xc:symbol-package 'fun-info)))
                              (symbolicate "FUN-INFO-" (second what)))
                           (fun-info-or-lose ',(first what)))
                          #',name))))))))
+
+(defmacro defoptimizers (kind names (lambda-list
+                                     &optional (node (gensym))
+                                     &rest vars)
+                         &body body)
+  (let ((optimizer-name (symbolicate (car names)
+                                     "-"
+                                     kind
+                                     "-OPTIMIZER")))
+    `(progn
+       (defoptimizer ,optimizer-name
+           (,lambda-list ,node ,@vars)
+         ,@body)
+       ,@(loop for name in names
+               collect `(setf (,(let ((*package* (sb-xc:symbol-package 'fun-info)))
+                                  (symbolicate "FUN-INFO-" kind))
+                               (fun-info-or-lose ',name))
+                              #',optimizer-name)))))
 
 ;;;; IR groveling macros
 
@@ -431,7 +466,7 @@
 (defmacro do-blocks ((block-var component &optional ends result) &body body)
   (unless (member ends '(nil :head :tail :both))
     (error "losing ENDS value: ~S" ends))
-  (let ((n-component (sb-xc:gensym))
+  (let ((n-component (gensym))
         (n-tail (gensym)))
     `(let* ((,n-component ,component)
             (,n-tail ,(if (member ends '(:both :tail))
@@ -447,7 +482,7 @@
 (defmacro do-blocks-backwards ((block-var component &optional ends result) &body body)
   (unless (member ends '(nil :head :tail :both))
     (error "losing ENDS value: ~S" ends))
-  (let ((n-component (sb-xc:gensym))
+  (let ((n-component (gensym))
         (n-head (gensym)))
     `(let* ((,n-component ,component)
             (,n-head ,(if (member ends '(:both :head))
@@ -582,21 +617,18 @@
 ;;; this table, as 42 is EQ to 42 no matter where in the source it
 ;;; appears. GET-SOURCE-PATH and NOTE-SOURCE-PATH functions should be
 ;;; always used to access this table.
-(declaim (hash-table *source-paths*))
+(declaim (type hash-table *source-paths*))
 (defvar *source-paths*)
 
 (defmacro with-source-paths (&body forms)
-  (with-unique-names (source-paths)
-    `(let* ((,source-paths (make-hash-table :test 'eq))
-            (*source-paths* ,source-paths))
-      (unwind-protect
-           (progn ,@forms)
-        (clrhash ,source-paths)))))
+  `(let ((*source-paths* (make-hash-table :test 'eq)))
+     ,@forms))
 
 ;;; Bind the hashtables used for keeping track of global variables,
 ;;; functions, etc.
 (defmacro with-ir1-namespace (&body forms)
-  `(let ((*ir1-namespace* (make-ir1-namespace))) ,@forms))
+  `(let ((*ir1-namespace* (make-ir1-namespace)))
+     ,@forms))
 
 ;;; Look up NAME in the lexical environment namespace designated by
 ;;; SLOT, returning the <value, T>, or <NIL, NIL> if no entry. The
@@ -623,6 +655,100 @@
            (setf (component-last-block ,component)
                  ,old-last-block))))))
 
+
+;;;; boolean attribute utilities
+;;;;
+;;;; We need to maintain various sets of boolean attributes for known
+;;;; functions and VOPs. To save space and allow for quick set
+;;;; operations, we represent the attributes as bits in a fixnum.
+
+(deftype attributes () 'fixnum)
+
+;;; Given a list of attribute names and an alist that translates them
+;;; to masks, return the OR of the masks.
+(defun encode-attribute-mask (names universe)
+  (loop for name in names
+        for pos = (position name universe)
+        sum (if pos (ash 1 pos) (error "unknown attribute name: ~S" name))))
+
+(defun decode-attribute-mask (bits universe)
+  (loop for name across universe
+        for mask = 1 then (ash mask 1)
+        when (logtest mask bits) collect name))
+
+;;; Define a new class of boolean attributes, with the attributes
+;;; having the specified ATTRIBUTE-NAMES. NAME is the name of the
+;;; class, which is used to generate some macros to manipulate sets of
+;;; the attributes:
+;;;
+;;;    NAME-attributep attributes attribute-name*
+;;;      Return true if one of the named attributes is present, false
+;;;      otherwise. When set with SETF, updates the place Attributes
+;;;      setting or clearing the specified attributes.
+;;;
+;;;    NAME-attributes attribute-name*
+;;;      Return a set of the named attributes.
+(defmacro !def-boolean-attribute (name &body attribute-names)
+  (let ((vector (coerce attribute-names 'vector))
+        (constructor (symbolicate name "-ATTRIBUTES"))
+        (test-name (symbolicate name "-ATTRIBUTEP")))
+    `(progn
+       (defmacro ,constructor (&rest attribute-names)
+         "Automagically generated boolean attribute creation function.
+  See !DEF-BOOLEAN-ATTRIBUTE."
+         (encode-attribute-mask attribute-names ,vector))
+       (defun ,(symbolicate "DECODE-" name "-ATTRIBUTES") (attributes)
+         (decode-attribute-mask attributes ,vector))
+       (defmacro ,test-name (attributes &rest attribute-names)
+         "Automagically generated boolean attribute test function.
+  See !DEF-BOOLEAN-ATTRIBUTE."
+         `(logtest (the attributes ,attributes)
+                   (,',constructor ,@attribute-names)))
+       (define-setf-expander ,test-name (place &rest attributes
+                                               &environment env)
+         "Automagically generated boolean attribute setter. See
+ !DEF-BOOLEAN-ATTRIBUTE."
+         (multiple-value-bind (temps values stores setter getter)
+             (#+sb-xc-host cl:get-setf-expansion
+              #-sb-xc-host get-setf-expansion place env)
+           (when (cdr stores)
+             (error "multiple store variables for ~S" place))
+           (let ((newval (gensym))
+                 (n-place (gensym))
+                 (mask (encode-attribute-mask attributes ,vector)))
+             (values `(,@temps ,n-place)
+                     `(,@values ,getter)
+                     `(,newval)
+                     `(let ((,(first stores)
+                             (if ,newval
+                                 (logior ,n-place ,mask)
+                                 (logandc2 ,n-place ,mask))))
+                        ,setter
+                        ,newval)
+                     `(,',test-name ,n-place ,@attributes))))))))
+
+;;; And now for some gratuitous pseudo-abstraction...
+;;;
+;;; ATTRIBUTES-UNION
+;;;   Return the union of all the sets of boolean attributes which are its
+;;;   arguments.
+;;; ATTRIBUTES-INTERSECTION
+;;;   Return the intersection of all the sets of boolean attributes which
+;;;   are its arguments.
+;;; ATTRIBUTES=
+;;;   True if the attributes present in ATTR1 are identical to
+;;;   those in ATTR2.
+(defmacro attributes-union (&rest attributes)
+  `(the attributes
+        (logior ,@(mapcar (lambda (x) `(the attributes ,x)) attributes))))
+(defmacro attributes-intersection (&rest attributes)
+  `(the attributes
+        (logand ,@(mapcar (lambda (x) `(the attributes ,x)) attributes))))
+(declaim (ftype (function (attributes attributes) boolean) attributes=))
+(declaim (inline attributes=))
+(defun attributes= (attr1 attr2)
+  (eql attr1 attr2))
+
 
 ;;;; the EVENT statistics/trace utility
 
@@ -844,3 +970,64 @@ specify bindings for printer control variables.")
         (nreverse (mapcar #'car *compiler-print-variable-alist*))
         (nreverse (mapcar #'cdr *compiler-print-variable-alist*))
       ,@forms)))
+
+#|
+In trying to figure out whether we can rectify the totally unobvious behavior
+that FUN-INFO-TRANSFORMS are tried in the reverse order of their definitions,
+it is helpful to understand when multiple transforms exist where more than
+one could be selected.
+In the absence of type overlap, it wouldn't matter what order they were attempted.
+Unfortunately there are plenty of cases where the _least_ _specific_ fun-type
+should be attempted first, which at least superfically makes no sense at all.
+
+(in-package sb-c)
+;; As a special case, delete any functions where there are 2 transforms
+;; and they can't possibly both apply. Just check the first arg.
+(defun domains-possibly-overlap (xforms)
+  (when (/= (length xforms) 2) ; lazy, just say yes
+    (return-from domains-possibly-overlap t))
+  (let ((t1 (transform-type (car xforms)))
+        (t2 (transform-type (cadr xforms))))
+    (unless (and (fun-type-p t1) (fun-type-p t2))
+      (return-from domains-possibly-overlap t))
+    (let ((req1 (car (fun-type-required t1)))
+          (req2 (car (fun-type-required t2))))
+      (unless (and req1 req2)
+        (return-from domains-possibly-overlap t))
+      (when (or (and (type= req1 (specifier-type 'single-float))
+                     (type= req2 (specifier-type 'double-float)))
+                (and (type= req1 (specifier-type 'double-float))
+                     (type= req2 (specifier-type 'single-float))))
+        (return-from domains-possibly-overlap nil))))
+  ;; The conservative answer is always T.
+  t)
+(let (list)
+  (do-all-symbols (s)
+    (dolist (name (list s `(setf ,s) `(cas ,s)))
+      (let* ((info (sb-int:info :function :info name))
+             (xforms (and info (sb-c::fun-info-transforms info))))
+        (when (and info
+                   (> (length xforms) 1)
+                   (not (assoc name list)))
+          (if (domains-possibly-overlap xforms)
+              (push (cons name xforms) list)
+              (format t "~&Nonoverlapping xforms on ~s~%" name))))))
+  (let ((*print-pretty* nil))
+    (dolist (x (sort (copy-list list) #'> :key (lambda (x) (length (cdr x))))
+               (format t "~s functions~%" (length list)))
+      (format t "~a~%" (car x))
+      (let ((i 0))
+        (dolist (xform (cdr x))
+          (incf i)
+          (format t " ~2d. ~s~%" i (sb-kernel:type-specifier
+                                  (sb-c::transform-type xform)))
+          (let* ((fun (sb-c::transform-function xform))
+                 (code (sb-kernel:fun-code-header fun))
+                 (info (sb-kernel:%code-debug-info code))
+                 (source (sb-c::compiled-debug-info-source info))
+                 (tlf (sb-c::compiled-debug-info-tlf-number info))
+                 (offs (sb-c::compiled-debug-info-char-offset info))
+                 (ns (sb-c::debug-source-namestring source))
+                 (doc (documentation (sb-c::transform-function xform) 'function)))
+            (format t "     ; ~a ~a ~d ~d~%" doc ns tlf offs)))))))
+|#
