@@ -443,7 +443,7 @@
 ;;;
 ;;; If the TN is an unused result TN, then we don't actually emit the
 ;;; move; we just change to the right kind of TN.
-(defun emit-coerce-vop (op dest-tn scs before)
+(defun emit-coerce-vop (op dest-tn scs before &optional load-scs)
   (declare (type tn-ref op) (type sc-vector scs) (type (or vop null) before)
            (type (or tn null) dest-tn))
   (let* ((op-tn (tn-ref-tn op))
@@ -467,7 +467,7 @@
                        (cond
                          ((not write-p)
                           (or
-                           (coerce-from-constant op temp)
+                           (coerce-from-constant op temp load-scs)
                            (emit-move (or (maybe-move-from-fixnum+-1 op-tn temp
                                                                      op)
                                           res)
@@ -499,26 +499,38 @@
 ;;; specified VOP. Dest-TN is the destination TN if we are doing a
 ;;; move or move-arg, and is NIL otherwise. This is only used for
 ;;; efficiency notes.
-(defun coerce-some-operands (ops dest-tn load-scs before)
+(defun coerce-some-operands (ops dest-tn load-scs before
+                             &optional more-load-scs)
   (declare (type (or tn-ref null) ops) (list load-scs)
            (type (or tn null) dest-tn) (type (or vop null) before))
   (do ((op ops (tn-ref-across op))
        (scs load-scs (cdr scs)))
-      ((null scs))
-    (let ((tn (tn-ref-tn op)))
+      ((null op))
+
+    (let* (more
+           (scs (cond (scs
+                       (car scs))
+                      (more-load-scs
+                       (setf more t)
+                       more-load-scs)
+                      (t
+                       (return))))
+           (tn (tn-ref-tn op)))
       (unless (or (eq (tn-kind tn) :unused)
-                  (svref (car scs)
+                  (svref scs
                          (sc-number (tn-sc tn))))
-        (emit-coerce-vop op dest-tn (car scs) before))))
+        (emit-coerce-vop op dest-tn scs before (and more
+                                                    more-load-scs)))))
   (values))
 
 ;;; Emit coerce VOPs for the args and results, as needed.
 (defun coerce-vop-operands (vop)
   (declare (type vop vop))
   (let ((info (vop-info vop)))
-    (coerce-some-operands (vop-args vop) nil (vop-info-arg-load-scs info) vop)
+    (coerce-some-operands (vop-args vop) nil (vop-info-arg-load-scs info) vop
+                          (vop-info-more-arg-load-scs info))
     (coerce-some-operands (vop-results vop) nil (vop-info-result-load-scs info)
-                          (vop-next vop)))
+                          (vop-next vop) nil))
   (values))
 
 ;;; Iterate over the more operands to a call VOP, emitting move-arg
@@ -557,21 +569,22 @@
 
         (change-tn-ref-tn val pass-tn)
         (let* ((this-fp
-                (cond ((not (sc-number-stack-p pass-sc)) fp-tn)
-                      (nfp-tn)
-                      (t
-                       (aver (eq how :known-return))
-                       (setq nfp-tn (make-number-stack-pointer-tn))
-                       (setf (tn-sc nfp-tn)
-                             (svref *backend-sc-numbers*
-                                    (first (primitive-type-scs
-                                            (tn-primitive-type nfp-tn)))))
-                       (emit-context-template
-                        node block
-                        (template-or-lose 'compute-old-nfp)
-                        nfp-tn vop)
-                       (aver (not (sc-number-stack-p (tn-sc nfp-tn))))
-                       nfp-tn)))
+                 (cond ((or (eq how :fixed)
+                            (not (sc-number-stack-p pass-sc))) fp-tn)
+                       (nfp-tn)
+                       (t
+                        (aver (eq how :known-return))
+                        (setq nfp-tn (make-number-stack-pointer-tn))
+                        (setf (tn-sc nfp-tn)
+                              (svref *backend-sc-numbers*
+                                     (first (primitive-type-scs
+                                             (tn-primitive-type nfp-tn)))))
+                        (emit-context-template
+                         node block
+                         (template-or-lose 'compute-old-nfp)
+                         nfp-tn vop)
+                        (aver (not (sc-number-stack-p (tn-sc nfp-tn))))
+                        nfp-tn)))
                (new (emit-move-arg-template node block res val-tn this-fp
                                             pass-tn vop))
                (after
@@ -583,7 +596,9 @@
                        (ir2-block-start-vop block)))))
           (coerce-some-operands (vop-args new) pass-tn
                                 (vop-info-arg-load-scs res)
-                                after)))))
+                                after)
+          (when (vop-info-after-sc-selection (vop-info new))
+            (funcall (vop-info-after-sc-selection (vop-info new)) new))))))
   (values))
 
 (defun maybe-move-from-fixnum+-1 (x y &optional x-tn-ref)
@@ -603,14 +618,21 @@
                                                   ,most-positive-fixnum)))
              (template-or-lose 'sb-vm::move-from-fixnum-1))))))
 
-(defun coerce-from-constant (x-tn-ref y)
+(defun coerce-from-constant (x-tn-ref y &optional load-scs)
   (when (and (sc-is y sb-vm::descriptor-reg sb-vm::control-stack)
              (tn-ref-type x-tn-ref))
     (multiple-value-bind (constantp value) (type-singleton-p (tn-ref-type x-tn-ref))
       (when constantp
-        (change-tn-ref-tn x-tn-ref
-                          (make-constant-tn (find-constant value) t))
-        t))))
+        (let* ((constant (find-constant value))
+               (sc (constant-sc constant)))
+          (when (or (not load-scs)
+                    ;; This is a more-arg-load-scs
+                    ;; Because more args do not have a generic load
+                    ;; sequence it can't handle aribtrary constants or
+                    ;; immediates.
+                    (svref load-scs (sc-number sc)))
+            (change-tn-ref-tn x-tn-ref (make-constant-tn constant t))
+            t))))))
 
 (defun split-ir2-block (vop)
   (cond ((vop-next vop)
@@ -622,6 +644,8 @@
                   (no-op-node (make-exit))
                   (new-2block (make-ir2-block block))
                   (vop-next (vop-next vop)))
+             (setf (block-number block)
+                   (incf (component-max-block-number (block-component block))))
              (link-node-to-previous-ctran no-op-node start)
              (setf (block-info block) new-2block)
              (add-to-emit-order new-2block (ir2-block-prev 2block))
@@ -673,7 +697,9 @@
                           (eq (car (vop-codegen-info last)) (ir2-block-%label dest-block))
                           (setf branch last)))))
       (let ((dest (tn-ref-tn (vop-results dest-vop))))
-        (when (and (eq (tn-sc x) (tn-sc dest))
+        (when (and (or (eq (tn-sc x) (tn-sc dest))
+                       (and (eq (sc-number (tn-sc x)) sb-vm:constant-sc-number)
+                            (eq (sc-number (tn-sc dest)) sb-vm:descriptor-reg-sc-number)))
                    (eq (find-move-vop x nil (tn-sc dest) (tn-primitive-type dest) #'sc-move-vops)
                        (vop-info vop)))
           (let ((new-block (split-ir2-block dest-vop))
@@ -755,12 +781,11 @@
 ;;; Arrange boxed constants so that all :NAMED-CALL constants are first,
 ;;; then constant leaves, and finally LOAD-TIME-VALUE constants.
 ;;; There exist a few reasons for placing all the FDEFNs first:
-;;;  * FDEFNs which are referenced for lisp call - as opposed to referenced
-;;;    in #'FUN syntax - could be stored as untagged pointers which would
-;;;    benefit the PPC64 architecture by removing a few instructions from each
-;;;    use of such fdefn by not having to subtract its lowtag prior to loading
-;;;    from both the fun and raw-fun slots. GC would need to be aware of the
-;;;    untagged pointer convention.
+;;;  * FDEFNs which are referenced for call (versus for value as in #'FUN)
+;;;    may be stored in the code header as untagged pointers. This benefits PPC64
+;;;    because lowtag subtraction can't be had "for free" due to the architectural
+;;;    requirement that lispword-aligned loads need a displacement that is
+;;;    a multiple of 4, which OTHER-POINTER-LOWTAG does not satisfy.
 ;;;  * In the current approach for so-called "static" linking of immobile code,
 ;;;    we change code instruction bytes so that they call into a simple-fun
 ;;;    directly rather than through an fdefn, but the approach is subject to a
@@ -787,11 +812,11 @@
              (dovector (constant unsorted)
                (incf old-offset)
                (when (eql pass (cond ((constant-p constant) 2)
-                                     ((eq (car constant) :named-call) 1)
+                                     ((eq (car constant) :fdefinition) 1)
                                      (t 3)))
                  (let ((new-offset (vector-push-extend constant sorted)))
                    (push (cons old-offset new-offset) renumbering))))))
-      (scan 1)  ; first all the called fdefinitions
+      (scan 1)  ; first all the fdefinitions use for call or global function ref (as in #'FUN)
       (scan 2)  ; then IR1 constants
       (scan 3)) ; then various flavors of load-time magic
     ;; Update the TN-OFFSET slot.
@@ -810,11 +835,14 @@
   (let (zero-tn)
     (flet ((zero-tn ()
              (or zero-tn
-                 (setf zero-tn
-                       (component-live-tn
-                        (make-wired-tn nil
-                                       sb-vm:any-reg-sc-number
-                                       sb-vm::zr-offset))))))
+                 (progn
+                   (setf zero-tn
+                         (component-live-tn
+                          (make-wired-tn nil
+                                         sb-vm:any-reg-sc-number
+                                         sb-vm::zr-offset))
+                         (tn-type zero-tn) (specifier-type '(eql 0)))
+                   zero-tn))))
       (do ((tn tn (tn-next tn)))
           ((null tn))
         (when (and (constant-tn-p tn)
@@ -871,13 +899,19 @@
                             (svref *backend-sc-numbers* sc)))
                      ;; Translate primitive type scs into constant scs
                      ((= sc sb-vm:descriptor-reg-sc-number)
-                      (setf (tn-sc tn)
-                            (svref *backend-sc-numbers* sb-vm:constant-sc-number)
-                            (tn-offset tn)
-                            (or (position (tn-leaf tn)
-                                          (ir2-component-constants 2comp))
-                                (vector-push-extend (tn-leaf tn)
-                                                    (ir2-component-constants 2comp)))))
+                      (cond #+(or arm64 x86-64)
+                            ((eql (tn-value tn) $0f0)
+                             ;; Can be loaded using just SINGLE-FLOAT-WIDETAG.
+                             (setf (tn-sc tn)
+                                   (svref *backend-sc-numbers* sb-vm:immediate-sc-number)))
+                            (t
+                             (setf (tn-sc tn)
+                                   (svref *backend-sc-numbers* sb-vm:constant-sc-number)
+                                   (tn-offset tn)
+                                   (or (position (tn-leaf tn)
+                                                 (ir2-component-constants 2comp))
+                                       (vector-push-extend (tn-leaf tn)
+                                                           (ir2-component-constants 2comp)))))))
                      (t
                       (setf (tn-sc tn)
                             (svref *backend-sc-numbers*
