@@ -374,6 +374,8 @@
   ;; collect dynamic statistics.)
   #+sb-dyncount
   (dyncount-info nil :type (or null dyncount-info))
+  ;; the number of jump table entries in this component
+  (n-jump-table-entries 0 :type index)
   ;; an array of references to lists of original source paths covered
   ;; for coverage instrumentation.
   (coverage-map (make-array 0 :fill-pointer 0 :adjustable t)
@@ -693,7 +695,7 @@
   ;;
   ;; :KNOWN-RETURN
   ;;     If needed, the old NFP is computed using COMPUTE-OLD-NFP.
-  (move-args nil :type (member nil :full-call :local-call :known-return))
+  (move-args nil :type (member nil :full-call :local-call :known-return :fixed))
   ;; a list of sc-vectors representing the loading costs of each fixed
   ;; argument and result
   (arg-costs nil :type list)
@@ -736,10 +738,12 @@
   ;; encodes the source ref (shifted 8, it is also encoded in
   ;; MAX-VOP-TN-REFS) and the dest ref index.
   (targets nil :type (or null (simple-array (unsigned-byte 16) 1)))
-  (optimizer nil :type (or null function))
+  (optimizer nil :type (or null function (cons function symbol)))
   (optional-results nil :type list)
   move-vop-p
   (after-sc-selection nil :type (or null function) :read-only t))
+(!set-load-form-method vop-info (:xc :target) :ignore-it)
+
 (declaim (inline vop-name))
 (defun vop-name (vop)
   (declare (type vop vop))
