@@ -17,30 +17,20 @@
 (defconstant sb-assem:+inst-alignment-bytes+ 1)
 
 (defconstant +backend-fasl-file-implementation+ :x86-64)
-(defconstant-eqx +fixup-kinds+ #(:absolute :relative :absolute64)
-  #'equalp)
+(defconstant-eqx +fixup-kinds+ #(:abs32 :rel32 :absolute) #'equalp)
 
 ;;; This size is supposed to indicate something about the actual granularity
 ;;; at which you can map memory.  We just hardwire it, but that may or may not
 ;;; be necessary any more.
 (defconstant +backend-page-bytes+ #+win32 65536 #-win32 32768)
 
-;;; The size in bytes of GENCGC cards, i.e. the granularity at which
-;;; writes to old generations are logged.
-;;; The size is a trade-off between efficiency of the allocator
-;;; and efficiency of scanning. Generally a card-marking GC will use cards
-;;; that are fairly small - The JVM used to use 512, I don't know if it still does.
-;;; I've heard of as small as 64 bytes being used in academic papers.
-;;; SBCL however has a problem with small sizes for two reasons:
-;;; (1) the size in which we claim memory in the slow-path allocator is exactly
-;;; the card size times a multiplier that does not work very well, and in fact
-;;; will work worse once I checkin a change to improve concurrency within the
-;;; slow path (2) heap scans can't begin at an arbitrary card because it might
-;;; the middle of a partly-boxed object. So we need to distinguish between
-;;; *strictly* boxed pages, and mixed-tagged/raw-word pages.
-;;; Alternatively, GC could try to skip over the cards at the start of a contiguous
-;;; block until it gets to the cards that are actully marked.
-(defconstant gencgc-page-bytes 16384)
+;;; The size in bytes of GENCGC pages. A page is the smallest amount of memory
+;;; that a thread can claim for a thread-local region, and also determines
+;;; the granularity at which we can find the start of a sequence of objects.
+(defconstant gencgc-page-bytes 32768)
+;;; The divisor relative to page-bytes which computes the granularity
+;;; at which writes to old generations are logged.
+(defconstant cards-per-page 32)
 ;;; The minimum size of new allocation regions.  While it doesn't
 ;;; currently make a lot of sense to have a card size lower than
 ;;; the alloc granularity, it will, once we are smarter about finding
@@ -106,7 +96,7 @@
                      :read-only-space-size #+metaspace #.(* 2 1024 1024)
                                            #-metaspace 0
                      :fixedobj-space-size #.(* 40 1024 1024)
-                     :varyobj-space-size #.(* 130 1024 1024)
+                     :text-space-size #.(* 130 1024 1024)
                      :dynamic-space-start #x1000000000)
 
 ;;; The default dynamic space size is lower on OpenBSD to allow SBCL to
@@ -114,12 +104,12 @@
 
 #-(or linux darwin)
 (!gencgc-space-setup #x20000000
-                     #-win32 :read-only-space-size #-win32 0
+                     :read-only-space-size 0
                      :dynamic-space-start #x1000000000
                      #+openbsd :dynamic-space-size #+openbsd #x1bcf0000)
 
-(defconstant linkage-table-growth-direction :up)
-(defconstant linkage-table-entry-size 16)
+(defconstant alien-linkage-table-growth-direction :up)
+(defconstant alien-linkage-table-entry-size 16)
 
 
 (defenum (:start 8)
@@ -166,33 +156,33 @@
      ;; interrupt handling
     #-sb-thread *pseudo-atomic-bits*     ; ditto
     #-sb-thread *binding-stack-pointer* ; ditto
+    ;; Since the text space and alien linkage table might both get relocated on startup
+    ;; under #+immobile-space, an alien callback wrapper can't wire in the address
+    ;; of a word that holds the C function pointer to callback_wrapper_trampoline.
+    ;; (There is no register that points to a known address when entering the callback)
+    ;; A static symbol works well for this, and is sensible considering that
+    ;; the assembled wrappers also reside in static space.
+    #+(and sb-thread immobile-space) callback-wrapper-trampoline
     *cpu-feature-bits*)
   #'equalp)
 
 (defconstant-eqx +static-fdefns+
-  #(length
-    two-arg-+
-    two-arg--
-    two-arg-*
-    two-arg-/
-    two-arg-<
-    two-arg->
-    two-arg-=
-    eql
-    %negate
-    two-arg-and
-    two-arg-ior
-    two-arg-xor
-    two-arg-gcd
-    two-arg-lcm
-    ensure-symbol-hash
-    sb-impl::install-hash-table-lock
-    update-object-layout
-    %coerce-callable-to-fun)
+    `#(ensure-symbol-hash sb-impl::install-hash-table-lock update-object-layout
+       ,@common-static-fdefns)
   #'equalp)
 
 #+sb-simd-pack
-(defglobal *simd-pack-element-types* '(integer single-float double-float))
+(defglobal *simd-pack-element-types*
+    '(single-float
+      double-float
+      (unsigned-byte 8)
+      (unsigned-byte 16)
+      (unsigned-byte 32)
+      (unsigned-byte 64)
+      (signed-byte 8)
+      (signed-byte 16)
+      (signed-byte 32)
+      (signed-byte 64)))
 
 (defconstant undefined-fdefn-header
   ;; This constant is constructed as follows: Take the INT opcode
