@@ -159,8 +159,7 @@
                          (convert-to-system-type type2))))))))
 
 (defun make-class-symbol (class-name)
-  (format-symbol #.(find-package "SB-PCL")
-                 "*THE-CLASS-~A*" (symbol-name class-name)))
+  (pcl-format-symbol "*THE-CLASS-~A*" (symbol-name class-name)))
 
 (defvar *standard-method-combination*)
 (defvar *or-method-combination*)
@@ -169,6 +168,10 @@
   (getf (object-plist object) name))
 
 (defun (setf plist-value) (new-value object name)
+  ;; HACK: exploit the fact we know the expansion of (SETF (GETF ...)) involves %PUTF.
+  ;; This keeps PCL plists in the system TLAB if this whole file is compiled in a global
+  ;; policy that does so. REMF is ok since it's a non-consing destructive operation.
+  (declare (inline sb-impl::%putf))
   (if new-value
       (setf (getf (object-plist object) name) new-value)
       (progn
@@ -201,7 +204,7 @@
                               (/noshow sub)
                               (when (member class (direct-supers sub) :test #'eq)
                                 (res sub)))))
-                        (res))))
+                        (sort (res) #'string< :key #'classoid-name))))
              `(list ,@(mapcar (lambda (kernel-bic-entry)
                                 (/noshow "setting up" kernel-bic-entry)
                                 (let* ((name (car kernel-bic-entry))
@@ -213,15 +216,15 @@
                                          ',(map 'list #'wrapper-classoid-name
                                                 (reverse (wrapper-inherits (classoid-wrapper class))))
                                          ,(getf (cdr kernel-bic-entry) :prototype-form))))
-                            (remove-if (lambda (kernel-bic-entry)
-                                         (member (first kernel-bic-entry)
-                                                 ;; remove special classes (T and our
-                                                 ;; SYSTEM-CLASSes) from the
-                                                 ;; BUILT-IN-CLASS list
-                                                 '(t function stream sequence
-                                                   file-stream string-stream
-                                                   slot-object)))
-                                       sb-kernel::*builtin-classoids*))))))
+                              (remove-if (lambda (kernel-bic-entry)
+                                           (member (first kernel-bic-entry)
+                                                   ;; remove special classes (T and our
+                                                   ;; SYSTEM-CLASSes) from the
+                                                   ;; BUILT-IN-CLASS list
+                                                   '(t function stream sequence
+                                                     file-stream string-stream
+                                                     slot-object)))
+                                         sb-kernel::*builtin-classoids*))))))
       (frob)))
 (/noshow "done setting up SB-PCL::*BUILT-IN-CLASSES*")
 
@@ -363,17 +366,22 @@
 (defclass method-combination (metaobject)
   ((%documentation :initform nil :initarg :documentation)))
 
-(defun make-gf-hash-table ()
+(defun make-gf-hashset ()
+  ;; Return what is logically a weak hashset, but physically a weak hash-table
+  ;; because we don't implement hashsets.
   (make-hash-table :test 'eq
                    :hash-function #'fsc-instance-hash ; stable hash
                    :weakness :key
                    :synchronized t))
+(defun add-to-weak-hashset (key set) (setf (gethash key set) t))
+(defun remove-from-weak-hashset (key set) (remhash key set))
+(defun weak-hashset-memberp (key set) (gethash key set))
 
 (defclass standard-method-combination (definition-source-mixin
                                        method-combination)
   ((type-name :reader method-combination-type-name :initarg :type-name)
    (options :reader method-combination-options :initarg :options)
-   (%generic-functions :initform (make-gf-hash-table)
+   (%generic-functions :initform (make-gf-hashset)
                        :reader method-combination-%generic-functions)))
 
 (defclass long-method-combination (standard-method-combination)
@@ -461,6 +469,11 @@
    (info
     :accessor slot-definition-info)))
 
+(defun uninitialized-accessor-function (type slotd)
+  (lambda (&rest args)
+    (declare (ignore args))
+    (error "~:(~A~) function~@[ for ~S ~] not yet initialized."
+           type slotd)))
 ;;; We use a structure here, because fast slot-accesses to this information
 ;;; are critical to making SLOT-VALUE-USING-CLASS &co fast: places that need
 ;;; these functions can access the SLOT-INFO directly, avoiding the overhead
