@@ -30,7 +30,6 @@
 #include "arch.h"
 #include "interr.h"
 #include "search.h"
-#include "purify.h"
 #include "globals.h"
 #include "lispregs.h"
 #include "interrupt.h"
@@ -39,7 +38,8 @@
 #include "genesis/primitive-objects.h"
 #include "genesis/gc-tables.h"
 #include "gc-internal.h"
-
+#include "tlsf-bsd/tlsf/tlsf.h"
+extern void* tlsf_control;
 
 /* When we need to do command input, we use this stream, which is not
  * in general stdin, so that things will "work" (as well as being
@@ -60,20 +60,26 @@ struct crash_preamble {
     uword_t signature;
     uword_t static_start;
     uword_t static_nbytes;
+    uword_t readonly_start;
+    uword_t readonly_nbytes;
     uword_t dynspace_start;
     long dynspace_npages;
     int card_size;
     int card_table_nbits;
     // fixedobj data dumped: pages, page table
     uword_t fixedobj_start, fixedobj_size, fixedobj_free_pointer;
-    // varyobj data dumped: pages, touched_bits, page table
-    uword_t varyobj_start, varyobj_size, varyobj_free_pointer;
+    // text data dumped: pages, touched_bits, page table
+    uword_t text_start, text_size;
+    lispobj *tlsf_mem_start, *text_space_highwatermark;
+    lispobj sentinel_block[3];
+    void* tlsf_control_address;
     int nthreads;
     int tls_size;
     lispobj lisp_package_vector;
+    int sizeof_context;
+    int tlsf_control_size;
     char sprof_enabled;
     char pin_dynspace_code;
-    int sizeof_context;
 };
 struct crash_thread_preamble {
     uword_t address;
@@ -86,10 +92,18 @@ struct crash_thread_preamble {
 const int CRASH_PREAMBLE_SIGNATURE =
     (sizeof (struct crash_preamble) << 16) | sizeof (struct crash_thread_preamble);
 
-#if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_SB_THREAD
+#if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64 // un-tested elsewhere
+#include <errno.h>
+static void checked_write(char *legend, int fd, void* buf, long nbytes)
+{
+    ssize_t wrote = write(fd, buf, nbytes);
+    if (wrote != nbytes) lose("short write, errno=%d", errno);
+    fprintf(stderr, "%s: %lx bytes\n", legend, nbytes);
+}
+
 #include "immobile-space.h"
 void save_gc_crashdump(char *pathname,
-                       uword_t approx_stackptr_at_gc_start)
+                       lispobj* cur_thread_approx_stackptr)
 {
     extern int pin_all_dynamic_space_code;
     int fd = open(pathname, O_WRONLY|O_CREAT, 0666);
@@ -99,11 +113,16 @@ void save_gc_crashdump(char *pathname,
     fprintf(stderr, "save: %d threads\n", nthreads);
     struct crash_preamble preamble;
     unsigned long nbytes_heap = next_free_page * GENCGC_PAGE_BYTES;
-    gc_assert(nbytes_heap == ((uword_t)dynamic_space_free_pointer - DYNAMIC_SPACE_START));
+#ifdef LISP_FEATURE_SB_THREAD
     int nbytes_tls = SymbolValue(FREE_TLS_INDEX,0);
+#else
+    int nbytes_tls = sizeof (struct thread);
+#endif
     preamble.signature = CRASH_PREAMBLE_SIGNATURE;
     preamble.static_start = STATIC_SPACE_START;
     preamble.static_nbytes = (uword_t)static_space_free_pointer - STATIC_SPACE_START;
+    preamble.readonly_start = READ_ONLY_SPACE_START;
+    preamble.readonly_nbytes = (uword_t)read_only_space_free_pointer - READ_ONLY_SPACE_START;
     preamble.dynspace_start = DYNAMIC_SPACE_START;
     preamble.dynspace_npages = next_free_page;
     preamble.card_size = GENCGC_CARD_BYTES;
@@ -115,64 +134,101 @@ void save_gc_crashdump(char *pathname,
     preamble.pin_dynspace_code = pin_all_dynamic_space_code;
     preamble.sizeof_context = sizeof (os_context_t);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
+    char *tlsf_memory_end = (char*)TEXT_SPACE_START + TEXT_SPACE_SIZE;
     preamble.fixedobj_start = FIXEDOBJ_SPACE_START;
     preamble.fixedobj_size = FIXEDOBJ_SPACE_SIZE;
     preamble.fixedobj_free_pointer = (uword_t)fixedobj_free_pointer;
-    preamble.varyobj_start = VARYOBJ_SPACE_START;
-    preamble.varyobj_size = VARYOBJ_SPACE_SIZE;
-    preamble.varyobj_free_pointer = (uword_t)varyobj_free_pointer;
+    preamble.text_start = TEXT_SPACE_START;
+    preamble.text_size = TEXT_SPACE_SIZE;
+    preamble.text_space_highwatermark = text_space_highwatermark;
+    preamble.tlsf_mem_start = tlsf_mem_start;
+    preamble.tlsf_control_address = tlsf_control;
+    preamble.tlsf_control_size = tlsf_size();
+    memcpy(preamble.sentinel_block, tlsf_memory_end-3*N_WORD_BYTES, 3*N_WORD_BYTES);
 #endif
-    // write the preamble and static space
-    write(fd, &preamble, sizeof preamble);
-    write(fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
+    // write the preamble and static + readonly spaces
+    checked_write("preamble", fd, &preamble, sizeof preamble);
+    checked_write("static", fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
+    checked_write("R/O", fd, (char*)READ_ONLY_SPACE_START, preamble.readonly_nbytes);
+
     // write the dynamic-space, PTEs, card table
-    write(fd, (char*)DYNAMIC_SPACE_START, nbytes_heap);
-    write(fd, page_table, sizeof (struct page) * next_free_page);
-    write(fd, gc_card_mark, 1+gc_card_table_mask);
+    checked_write("dynamic", fd, (char*)DYNAMIC_SPACE_START, nbytes_heap);
+    checked_write("PTE", fd, page_table, sizeof (struct page) * next_free_page);
+    checked_write("cardmark", fd, gc_card_mark, 1+gc_card_table_mask);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     int usage = (uword_t)fixedobj_free_pointer - FIXEDOBJ_SPACE_START;
-    write(fd, (char*)FIXEDOBJ_SPACE_START, usage);
+    checked_write("fixedobj", fd, (char*)FIXEDOBJ_SPACE_START, usage);
     int total_npages = FIXEDOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
-    write(fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
-    usage = (uword_t)varyobj_free_pointer - VARYOBJ_SPACE_START;
-    write(fd, (char*)VARYOBJ_SPACE_START, usage);
-    total_npages = VARYOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
+    checked_write("fixedobj_PTE", fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
+    usage = (uword_t)text_space_highwatermark - TEXT_SPACE_START;
+    // write the block_header_t that is just beyond the high water mark
+    checked_write("text", fd, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
+    total_npages = TEXT_SPACE_SIZE / IMMOBILE_CARD_BYTES;
     int n_bitmap_elts = ALIGN_UP(total_npages, 32) / 32;
-    write(fd, varyobj_page_touched_bits, n_bitmap_elts * sizeof (int));
-    write(fd, varyobj_pages, total_npages * sizeof (int));
+    checked_write("text_gen", fd, text_page_genmask, total_npages); // 1 byte per page
+    checked_write("text_WP", fd, text_page_touched_bits, n_bitmap_elts * sizeof (int));
+    checked_write("TLSF_control", fd, tlsf_control, preamble.tlsf_control_size);
+    int tlsf_memory_size = tlsf_memory_end - (char*)tlsf_mem_start;
+    int n_tlsf_pages = tlsf_memory_size / IMMOBILE_CARD_BYTES;
+    checked_write("TLSF_sso", fd, tlsf_page_sso, n_tlsf_pages * sizeof (short));
 #endif
     struct crash_thread_preamble thread_preamble;
     for_each_thread(th) {
         int ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,th));
         os_context_t* threadcontext = nth_interrupt_context(0, th);
         uword_t sp;
-        if (ici)
+        if (ici) {
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
             sp = *os_context_register_addr(threadcontext, reg_SP);
-        else {
+#else
+            sp = *os_context_register_addr(threadcontext, reg_CSP);
+#endif
+        } else if (th->state_word.state == STATE_DEAD) {
+            gc_assert(th->binding_stack_pointer == th->binding_stack_start);
+            // FIXME: assumes stack grows down
+            sp = (uword_t)th->control_stack_end;
+        } else {
             if (th != get_sb_vm_thread()) {
-              char msg[] = "No stackptr for crash dump\n";
-              write(2, msg, sizeof msg-1);
+              char msg[80];
+              int n = snprintf(msg, sizeof msg,
+                               "thread %p state %d - No stackptr for crash dump\n",
+                               th, th->state_word.state);
+              write(2, msg, n);
               _exit(1);
             }
-            sp = approx_stackptr_at_gc_start;
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
+            sp = (uword_t)cur_thread_approx_stackptr;
+#else
+            sp = access_control_stack_pointer(th);
+#endif
         }
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
         int nbytes_control_stack =
             (char*)th->control_stack_end - (char*)sp; // grows downward
         int nbytes_binding_stack =
             (char*)th->binding_stack_pointer - (char*)th->binding_stack_start; // grows upward
+#else
+        int nbytes_control_stack = (char*)sp - (char*)th->control_stack_start; // grows upward
+        int nbytes_binding_stack =
+          (char*)get_binding_stack_pointer(th) - (char*)th->binding_stack_start;
+#endif
         thread_preamble.address = (uword_t)th;
         thread_preamble.has_context = ici != 0; // boolean for have context or not
         thread_preamble.control_stack_nbytes = nbytes_control_stack;
         thread_preamble.binding_stack_nbytes = nbytes_binding_stack;
         // write the preamble
-        write(fd, &thread_preamble, sizeof thread_preamble);
+        checked_write("thread", fd, &thread_preamble, sizeof thread_preamble);
         // write 0 or 1 contexts, control-stack, binding-stack, TLS
-        if (ici) write(fd, threadcontext, preamble.sizeof_context);
-        write(fd, (char*)sp, nbytes_control_stack);
-        write(fd, th->binding_stack_start, nbytes_binding_stack);
-        write(fd, th, nbytes_tls);
+        if (ici) checked_write(" ctxt", fd, threadcontext, preamble.sizeof_context);
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
+        checked_write(" stack", fd, (char*)sp, nbytes_control_stack);
+#else
+        checked_write(" stack", fd, th->control_stack_start, nbytes_control_stack);
+#endif
+        checked_write(" bindings", fd, th->binding_stack_start, nbytes_binding_stack);
+        checked_write(" TLS", fd, th, nbytes_tls);
     }
-    write(fd, "SB.Crash", 8); // trailing signature
+    checked_write("sig", fd, "SB.Crash", 8); // trailing signature
     close(fd);
 }
 #endif
@@ -180,36 +236,69 @@ void save_gc_crashdump(char *pathname,
 static cmd call_cmd, dump_cmd, print_cmd, quit_cmd, help_cmd;
 static cmd flush_cmd, regs_cmd, exit_cmd;
 static cmd print_context_cmd, pte_cmd, search_cmd;
-static cmd backtrace_cmd, purify_cmd, catchers_cmd;
-static cmd grab_sigs_cmd;
-static cmd kill_cmd;
+static cmd backtrace_cmd, catchers_cmd;
+static cmd threads_cmd;
+
+extern void gc_stop_the_world(), gc_start_the_world();
+static void suspend_other_threads() {
+    gc_stop_the_world();
+    // It might make sense for each thread's stop-for-gc handler to close its region
+    // versus doing this loop
+    struct thread *th;
+    for_each_thread(th) { gc_close_thread_regions(th, 0); }
+    gc_close_collector_regions(0);
+}
+static void unsuspend_other_threads() {
+    gc_start_the_world();
+}
 
 static int save_cmd(char **ptr) {
-#if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_SB_THREAD
-    extern void gc_stop_the_world(), gc_start_the_world();
     char *name  = parse_token(ptr);
     if (!name) {
         fprintf(stderr, "Need filename\n");
         return 1;
     }
-    gc_stop_the_world();
-    // It might make sense for each thread's stop-for-gc handler to close its region
-    // versus doing this loop
-    struct thread *th;
-    for_each_thread(th) { gc_close_thread_regions(th); }
-    gc_close_collector_regions();
-    save_gc_crashdump(name, (uword_t)__builtin_frame_address(0));
-    gc_start_the_world();
+#if (defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64) && defined LISP_FEATURE_SB_THREAD
+    suspend_other_threads();
+    save_gc_crashdump(name, (lispobj*)__builtin_frame_address(0));
+    unsuspend_other_threads();
 #else
     fprintf(stderr, "Unimplemented\n");
 #endif
     return 0;
 }
-
-#ifdef STANDALONE
-static int verify_cmd(char **ptr) {
+void list_lisp_threads(int regions) {
+    struct thread* th;
+    fprintf(stderr, "(thread*,pthread,sb-vm:thread,name)\n");
+    void* pthread;
+    for_each_thread(th) {
+        memcpy(&pthread, &th->os_thread, N_WORD_BYTES);
+        struct thread_instance* i = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
+        char* name = NULL;
+        if (i->name != NIL &&
+            widetag_of(native_pointer(i->name)) == SIMPLE_BASE_STRING_WIDETAG)
+            name = (char*)VECTOR(i->name)->data;
+        fprintf(stderr, "%p %p %p \"%s\"\n", th, pthread, (void*)i, name);
+        if (regions) {
+#define show_tlab(label, r) fprintf(stderr, "  %s @ %p: %p %p %p\n", label, \
+         &th->r, th->r.start_addr, th->r.end_addr, th->r.free_pointer)
+            show_tlab("usr cons ", cons_tlab);
+            show_tlab("usr mix  ", mixed_tlab);
+            show_tlab("sys cons ", sys_cons_tlab);
+            show_tlab("sys mix  ", sys_mixed_tlab);
+#undef show_tlab
+        }
+    }
+}
+static int threads_cmd(char **ptr) {
+    list_lisp_threads(more_p(ptr) && !strncmp(*ptr, "-r", 2));
+    return 0;
+}
+static int verify_cmd(char __attribute__((unused)) **ptr) {
     gencgc_verbose = 1;
-    verify_heap(0);
+    suspend_other_threads();
+    verify_heap(0, 0);
+    unsuspend_other_threads();
     return 0;
 }
 static int gc_cmd(char **ptr) {
@@ -220,10 +309,18 @@ static int gc_cmd(char **ptr) {
     gencgc_verbose = 2;
     pre_verify_gen_0 = 1;
     verify_gens = 0;
+    suspend_other_threads();
     collect_garbage(last_gen);
+    unsuspend_other_threads();
     return 0;
 }
+
+static int tlsf_cmd(__attribute__((unused)) char **ptr) {
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    tlsf_dump_pool(tlsf_control, tlsf_mem_start, "/dev/tty");
 #endif
+    return 0;
+}
 
 static struct cmd {
     char *cmd, *help;
@@ -239,13 +336,6 @@ static struct cmd {
     {"d", "(an alias for dump)", dump_cmd},
     {"exit", "Exit this instance of the monitor.", exit_cmd},
     {"flush", "Flush all temp variables.", flush_cmd},
-    /* (Classic CMU CL had a "gc" command here, which seems like a
-     * reasonable idea, but the code was stale (incompatible with
-     * gencgc) so I just flushed it. -- WHN 20000814 */
-    {"grab-signals", "Set the signal handlers to call LDB.", grab_sigs_cmd},
-    {"kill", "Kill ourself with signal number N (useful if running under gdb)",
-     kill_cmd},
-    {"purify", "Purify. (Caveat purifier!)", purify_cmd},
     {"print", "Print object at ADDRESS.", print_cmd},
     {"p", "(an alias for print)", print_cmd},
     {"pte", "Page table entry for address", pte_cmd},
@@ -253,10 +343,10 @@ static struct cmd {
     {"regs", "Display current Lisp registers.", regs_cmd},
     {"search", "Search heap for object.", search_cmd},
     {"save", "Produce crashdump", save_cmd},
-#ifdef STANDALONE
+    {"threads", "List threads", threads_cmd},
+    {"tlsfdump", "Dump TLSF structures", tlsf_cmd},
     {"verify", "Check heap invariants", verify_cmd},
     {"gc", "Collect garbage", gc_cmd},
-#endif
     {NULL, NULL, NULL}
 };
 
@@ -285,14 +375,14 @@ dump_cmd(char **ptr)
     int force = 0, decode = 0;
 
     if (more_p(ptr)) {
-        // you can't both "force" and "decode" - only one or the other,
-        // or neither
-        if (!strncmp(*ptr, "-f ", 3)) {
-          force = 1;
-          *ptr += 3;
-        } else if (!strncmp(*ptr, "-d ", 3)) {
-          decode = 1;
-          *ptr += 3;
+        while (1) {
+            if (!strncmp(*ptr, "-f ", 3)) {
+              force = 1;
+              *ptr += 3;
+            } else if (!strncmp(*ptr, "-d ", 3)) {
+              decode = 1;
+              *ptr += 3;
+            } else break;
         }
         if (!parse_addr(ptr, !force, &addr)) return 0;
 
@@ -360,7 +450,7 @@ dump_cmd(char **ptr)
                 if (word != 0 && !is_lisp_pointer(word)
                     && valid_widetag_p(header_widetag(word))) {
                     printf(" %s", widetag_names[header_widetag(word)>>2]);
-                    next_object += sizetab[header_widetag(word)](next_object);
+                    next_object += headerobj_size2(next_object, word);
                 } else if (!is_header(word)) {
                     next_object += CONS_SIZE;
                 } else { // disable decoder if weirdness observed
@@ -396,16 +486,6 @@ pte_cmd(char **ptr)
     return 0;
 }
 
-static int
-kill_cmd(char **ptr)
-{
-#ifndef LISP_FEATURE_WIN32
-    int sig;
-    if (parse_number(ptr, &sig)) kill(getpid(), sig);
-#endif
-    return 0;
-}
-
 static int
 regs_cmd(char __attribute__((unused)) **ptr)
 {
@@ -431,12 +511,6 @@ regs_cmd(char __attribute__((unused)) **ptr)
     printf("DYNAMIC\t=\t%p\n", (void*)current_dynamic_space);
 #endif
 
-#ifndef ALLOCATION_POINTER
-    printf("ALLOC\t=\t%p\n", (void*)dynamic_space_free_pointer);
-#else
-    printf("ALLOC\t=\t%p\n", (void*)SymbolValue(ALLOCATION_POINTER, thread));
-#endif
-
 #ifndef LISP_FEATURE_GENCGC
     printf("TRIGGER\t=\t%p\n", (void*)current_auto_gc_trigger);
 #endif
@@ -556,13 +630,6 @@ exit_cmd(char __attribute__((unused)) **ptr)
     return 1; // 'done' flag
 }
 
-static int
-purify_cmd(char __attribute__((unused)) **ptr)
-{
-    purify(NIL, NIL);
-    return 0;
-}
-
 static void
 print_context(os_context_t *context)
 {
@@ -573,10 +640,6 @@ print_context(os_context_t *context)
         brief_print((lispobj)(*os_context_register_addr(context,i)));
 
     }
-#if defined(LISP_FEATURE_DARWIN) && defined(LISP_FEATURE_PPC)
-    printf("DAR:\t\t 0x%08lx\n", (unsigned long)(*os_context_register_addr(context, 41)));
-    printf("DSISR:\t\t 0x%08lx\n", (unsigned long)(*os_context_register_addr(context, 42)));
-#endif
 #ifndef REG_PC
     printf("PC:\t\t  0x%08lx\n", (unsigned long)os_context_pc(context));
 #endif
@@ -600,12 +663,11 @@ print_context_cmd(char **ptr)
             printf("printing context %d\n", index);
             print_context(nth_interrupt_context(index, thread));
         } else {
-            printf("There aren't that many/few contexts.\n");
             printf("There are %d interrupt contexts.\n", free_ici);
         }
     } else {
         if (free_ici == 0)
-            printf("There are no interrupt contexts!\n");
+            printf("There are no interrupt contexts.\n");
         else {
             printf("There are %d interrupt contexts.\n", free_ici);
             printf("printing context %d\n", free_ici - 1);
@@ -661,7 +723,7 @@ catchers_cmd(char __attribute__((unused)) **ptr)
 #if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64) || defined(LISP_FEATURE_ARM64)
                    component_ptr_from_pc((void*)catch->entry_pc),
 #else
-                   catch->code,
+                   (void*)catch->code,
 #endif
                    (void*)(catch->entry_pc));
             brief_print((lispobj)catch->tag);
@@ -671,34 +733,6 @@ catchers_cmd(char __attribute__((unused)) **ptr)
     return 0;
 }
 
-/* SIGINT handler that invokes the monitor (for when Lisp isn't up to it) */
-static void
-sigint_handler(int __attribute__((unused)) signal,
-               siginfo_t __attribute__((unused)) *info,
-               void *context)
-{
-    extern void ldb_monitor();
-    fprintf(stderr, "\nSIGINT hit at %p\n", (void*)os_context_pc(context));
-    ldb_monitor();
-    fprintf(stderr, "Returning to lisp (if you're lucky).\n");
-}
-
-static int
-grab_sigs_cmd(char __attribute__((unused)) **ptr)
-{
-#ifdef LISP_FEATURE_WIN32
-    fprintf(stderr, "sorry no can do\n"); fflush(stderr);
-#else
-    printf("Grabbing SIGINT.\n");
-    struct sigaction sa;
-    sigemptyset(&sa.sa_mask);
-    sa.sa_flags = SA_SIGINFO;
-    sa.sa_sigaction = sigint_handler;
-    sigaction(SIGINT, &sa, 0);
-#endif
-    return 0;
-}
-
 extern boolean gc_active_p;
 extern FILE *gc_activitylog_file;
 void
@@ -710,7 +744,8 @@ ldb_monitor(void)
     int ambig;
 
     printf("Welcome to LDB, a low-level debugger for the Lisp runtime environment.\n");
-    if (gc_active_p) printf("(GC in progress)\n");
+    if (gc_active_p) printf("(GC in progress, oldspace=%d, newspace=%d)\n",
+                            from_space, new_space);
     if (gc_activitylog_file) fflush(gc_activitylog_file);
     if (!ldb_in) {
 #ifndef LISP_FEATURE_WIN32
@@ -769,32 +804,41 @@ monitor_or_something()
     ldb_monitor();
 }
 
-#ifdef STANDALONE
+#ifdef STANDALONE_LDB
+void gc_stop_the_world() { } // do nothing
+void gc_start_the_world() { } // do nothing
 #include <errno.h>
 #include "core.h"
 #include "gencgc-private.h"
 struct lisp_startup_options lisp_startup_options;
 
 void unwind_binding_stack() { lose("Can't unwind binding stack"); }
-FILE *prepare_to_save(char *filename, boolean prepend_runtime, void **runtime_bytes,
-                      size_t *runtime_size) {
+FILE *prepare_to_save(__attribute__((unused)) char *filename,
+                      __attribute__((unused)) boolean prepend_runtime,
+                      __attribute__((unused)) void **runtime_bytes,
+                      __attribute__((unused)) size_t *runtime_size) {
     lose("Can't prepare_to_save");
 }
-boolean save_runtime_to_filehandle(FILE *output, void *runtime, size_t runtime_size,
-                                   int application_type) {
+boolean save_runtime_to_filehandle(__attribute__((unused)) FILE *output,
+                                   __attribute__((unused)) void *runtime,
+                                   __attribute__((unused)) size_t runtime_size,
+                                   __attribute__((unused)) int application_type) {
     lose("Can't save_runtime_to_filehandle");
 }
-boolean save_to_filehandle(FILE *file, char *filename, lispobj init_function,
-                           boolean make_executable,
-                           boolean save_runtime_options,
-                           int core_compression_level) {
+boolean save_to_filehandle(__attribute__((unused)) FILE *file,
+                           __attribute__((unused)) char *filename,
+                           __attribute__((unused)) lispobj init_function,
+                           __attribute__((unused)) boolean make_executable,
+                           __attribute__((unused)) boolean save_runtime_options,
+                           __attribute__((unused)) int core_compression_level) {
     lose("Can't save_to_filehandle");
 }
 
-static size_t checked_read(int fd, void* buf, size_t n)
+static size_t checked_read(char *legend, int fd, void* buf, size_t n)
 {
     size_t result = read(fd, buf, n);
     if (result != n) { lose("read failed, errno=%d", errno); }
+    if (legend) fprintf(stderr, "%s: %lx bytes\n", legend, n);
     return result;
 }
 
@@ -813,7 +857,7 @@ char *pagetypedesc(int type)
 extern void gc_allocate_ptes();
 extern void recompute_gen_bytes_allocated();
 extern void print_generation_stats();
-extern struct thread *alloc_thread_struct(void*,lispobj);
+extern struct thread *alloc_thread_struct(void*);
 
 int load_gc_crashdump(char* pathname)
 {
@@ -828,94 +872,125 @@ int load_gc_crashdump(char* pathname)
     }
     struct crash_preamble preamble;
     struct crash_thread_preamble thread_preamble;
-    checked_read(fd, &preamble, sizeof preamble);
-    printf("static==%lx size=%x\n", preamble.static_start, (int)preamble.static_nbytes);
-    printf("heap_start=%lx npages=%d\n", preamble.dynspace_start, (int)preamble.dynspace_npages);
+    checked_read("preamble", fd, &preamble, sizeof preamble);
+    printf("static=%"OBJ_FMTX" nbytes=%x\n", preamble.static_start, (int)preamble.static_nbytes);
+    printf("heap_start=%"OBJ_FMTX" npages=%d\n", preamble.dynspace_start, (int)preamble.dynspace_npages);
     // pin_dynspace_code is for display only. It gets recomputed as the
     // logical OR of all threads' values of *GC-PIN-CODE-PAGES*.
     printf("sprof_enabled=%d pin_dynspace_code=%d packages=%p\n",
            preamble.sprof_enabled, preamble.pin_dynspace_code,
            (void*)preamble.lisp_package_vector);
+    lisp_package_vector = preamble.lisp_package_vector;
     sb_sprof_enabled = preamble.sprof_enabled;
     if (preamble.signature != CRASH_PREAMBLE_SIGNATURE)
-        lose("Can't load crashdump: bad header (have %lx, expect %lx)",
-             preamble.signature, (long)CRASH_PREAMBLE_SIGNATURE);
+        lose("Can't load crashdump: bad header (have %"OBJ_FMTX", expect %"OBJ_FMTX")",
+             preamble.signature, (uword_t)CRASH_PREAMBLE_SIGNATURE);
     if (preamble.card_size != GENCGC_CARD_BYTES)
         lose("Can't load crashdump: memory parameters differ");
     gc_card_table_nbits = preamble.card_table_nbits;
     gc_allocate_ptes();
     next_free_page = preamble.dynspace_npages;
-    checked_read(fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
+    // static + readonly
+    checked_read("static", fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
     static_space_free_pointer = (lispobj*)(STATIC_SPACE_START + preamble.static_nbytes);
+    if (preamble.readonly_nbytes) {
+      // READ_ONLY_SPACE_START = preamble.readonly_start;
+      os_alloc_gc_space(READ_ONLY_CORE_SPACE_ID, 0, (char*)preamble.readonly_start,
+                        ALIGN_UP(preamble.readonly_nbytes, 4096));
+      checked_read("R/O", fd, (char*)preamble.readonly_start, preamble.readonly_nbytes);
+    }
+    read_only_space_free_pointer = (lispobj*)(READ_ONLY_SPACE_START + preamble.readonly_nbytes);
+    //
     DYNAMIC_SPACE_START = preamble.dynspace_start;
     long dynspace_nbytes = preamble.dynspace_npages * GENCGC_PAGE_BYTES;
-    dynamic_space_free_pointer = (void*)(DYNAMIC_SPACE_START + dynspace_nbytes);
-    char* dynspace = os_validate(0, (char*)preamble.dynspace_start,
-                                 DEFAULT_DYNAMIC_SPACE_SIZE, 0, 0);
+    char* dynspace = os_alloc_gc_space(DYNAMIC_CORE_SPACE_ID, 0, (char*)preamble.dynspace_start,
+                                       DEFAULT_DYNAMIC_SPACE_SIZE);
     if (dynspace != (char*)preamble.dynspace_start)
         lose("Didn't map dynamic space where expected: %p vs %p",
              dynspace, (char*)preamble.dynspace_start);
-    checked_read(fd, (char*)DYNAMIC_SPACE_START, dynspace_nbytes);
-    fprintf(stderr, "snapshot: %ld pages in use (%ld bytes)\n",
+    checked_read("dynamic", fd, (char*)DYNAMIC_SPACE_START, dynspace_nbytes);
+    fprintf(stderr, "snapshot: %"PRIdPTR" pages in use (%ld bytes)\n",
             next_free_page, dynspace_nbytes);
-    checked_read(fd, page_table, sizeof (struct page) * next_free_page);
+    checked_read("PTE", fd, page_table, sizeof (struct page) * next_free_page);
     recompute_gen_bytes_allocated();
-    checked_read(fd, gc_card_mark, 1+gc_card_table_mask);
+    checked_read("cardmark", fd, gc_card_mark, 1+gc_card_table_mask);
     print_generation_stats();
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     extern void gc_init_immobile(),
                 calc_immobile_space_bounds(),
                 write_protect_immobile_space();
     gc_assert(preamble.fixedobj_size == FIXEDOBJ_SPACE_SIZE);
-    gc_assert(preamble.varyobj_size == VARYOBJ_SPACE_SIZE);
+    gc_assert(preamble.text_size == TEXT_SPACE_SIZE);
     FIXEDOBJ_SPACE_START = preamble.fixedobj_start;
-    VARYOBJ_SPACE_START = preamble.varyobj_start;
+    TEXT_SPACE_START = preamble.text_start;
     fixedobj_free_pointer = (lispobj*)preamble.fixedobj_free_pointer;
-    varyobj_free_pointer = (lispobj*)preamble.varyobj_free_pointer;
-    os_validate(0, (char*)FIXEDOBJ_SPACE_START, FIXEDOBJ_SPACE_SIZE, 0, 0);
-    os_validate(0, (char*)VARYOBJ_SPACE_START, VARYOBJ_SPACE_SIZE, 0, 0);
+    text_space_highwatermark = (lispobj*)preamble.text_space_highwatermark;
+    os_alloc_gc_space(IMMOBILE_FIXEDOBJ_CORE_SPACE_ID, 0, (char*)FIXEDOBJ_SPACE_START,
+                      FIXEDOBJ_SPACE_SIZE);
+    os_alloc_gc_space(IMMOBILE_TEXT_CORE_SPACE_ID, 0, (char*)TEXT_SPACE_START,
+                      TEXT_SPACE_SIZE);
     gc_init_immobile(); // allocate the page tables
     calc_immobile_space_bounds();
     // Read fixedobj space
     int usage = (uword_t)fixedobj_free_pointer - FIXEDOBJ_SPACE_START;
-    checked_read(fd, (char*)FIXEDOBJ_SPACE_START, usage);
+    checked_read("fixedobj", fd, (char*)FIXEDOBJ_SPACE_START, usage);
     // Always read the whole page table regardless of the current space usage
     int total_npages = FIXEDOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
-    checked_read(fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
-    // Read varyobj space
-    usage = (uword_t)varyobj_free_pointer - VARYOBJ_SPACE_START;
-    checked_read(fd, (char*)VARYOBJ_SPACE_START, usage);
-    total_npages = VARYOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
+    checked_read("fixedobj_PTE", fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
+    // Read text space
+    usage = (uword_t)text_space_highwatermark - TEXT_SPACE_START;
+    char *tlsf_memory_end = (char*)TEXT_SPACE_START + TEXT_SPACE_SIZE;
+    tlsf_mem_start = preamble.tlsf_mem_start;
+    fprintf(stderr, "tlsf_mem_start=%p\n", tlsf_mem_start);
+    int tlsf_memory_size = tlsf_memory_end - (char*)tlsf_mem_start;
+    checked_read("text", fd, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
+    memcpy(tlsf_memory_end-3*N_WORD_BYTES, preamble.sentinel_block, 3*N_WORD_BYTES);
+    total_npages = TEXT_SPACE_SIZE / IMMOBILE_CARD_BYTES;
     int n_bitmap_elts = ALIGN_UP(total_npages, 32) / 32;
-    checked_read(fd, varyobj_page_touched_bits, n_bitmap_elts * sizeof (int));
-    checked_read(fd, varyobj_pages, total_npages * sizeof (int));
+    checked_read("text_gen", fd, text_page_genmask, total_npages); // 1 byte per page
+    checked_read("text_WP", fd, text_page_touched_bits, n_bitmap_elts * sizeof (int));
+    tlsf_control = preamble.tlsf_control_address; // already mapped at a fixed address
+    // TLSF control was mapped in gc_init_immobile()
+    checked_read("TLSF_control", fd, tlsf_control, preamble.tlsf_control_size);
+    int n_tlsf_pages = tlsf_memory_size / IMMOBILE_CARD_BYTES;
+    fprintf(stderr, "%d TLSF pages\n", n_tlsf_pages);
+    tlsf_page_sso = malloc(n_tlsf_pages * sizeof (short));
+    checked_read("TLSF_sso", fd, tlsf_page_sso, n_tlsf_pages * sizeof (short));
     write_protect_immobile_space();
 #endif
     fprintf(stderr, "%d threads:\n", (int)preamble.nthreads);
     int i;
     for(i=0; i<(int)preamble.nthreads; ++i) {
-        struct thread* th = alloc_thread_struct(0, 0);
+        struct thread* th = alloc_thread_struct(0);
         // Push it on the front
         th->prev = 0;
         th->next = threads;
         if (threads) threads->prev = th;
         threads = th;
-        checked_read(fd, &thread_preamble, sizeof thread_preamble);
+        checked_read("thread", fd, &thread_preamble, sizeof thread_preamble);
         uword_t* stackptr = (uword_t*)((char*)th->control_stack_end - thread_preamble.control_stack_nbytes);
         context = contexts[i] = malloc(preamble.sizeof_context);
         nth_interrupt_context(0, th) = context;
         if (thread_preamble.has_context) {
-            checked_read(fd, context, preamble.sizeof_context);
+            checked_read(" ctxt", fd, context, preamble.sizeof_context);
         }
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
         *os_context_sp_addr(context) = (uword_t)stackptr;
-        checked_read(fd, stackptr, thread_preamble.control_stack_nbytes);
-        checked_read(fd, th->binding_stack_start, thread_preamble.binding_stack_nbytes);
+#ifdef LISP_FEATURE_X86
+        *os_context_register_addr(context,reg_SP) = (os_context_register_t)stackptr;
+#endif
+        gc_assert(*os_context_register_addr(context,reg_SP) == (os_context_register_t)stackptr);
+#else
+        *os_context_register_addr(context, reg_CSP) = (uword_t)stackptr;
+#endif
+        checked_read(" stack", fd, stackptr, thread_preamble.control_stack_nbytes);
+        checked_read(" bindings", fd, th->binding_stack_start, thread_preamble.binding_stack_nbytes);
         // Skip over the initial words of the thread structure that was saved
         // in the file, so that binding_stack_start remains as is in the
         // newly allocated structure. The last word is the only one we want to keep.
         int skip = sizeof dummy-N_WORD_BYTES;
-        checked_read(fd, &dummy, skip);
-        checked_read(fd, &th->lisp_thread, preamble.tls_size-skip);
+        checked_read(0, fd, &dummy, skip);
+        checked_read(" TLS", fd, &th->lisp_thread, preamble.tls_size-skip);
         write_TLS(FREE_INTERRUPT_CONTEXT_INDEX, make_fixnum(1), th);
         struct thread_instance* instance = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
         lispobj name = instance->name;
@@ -959,7 +1034,7 @@ int load_gc_crashdump(char* pathname)
         putc('\n', stderr);
     }
     char signature[8];
-    checked_read(fd, signature, 8);
+    checked_read("sig", fd, signature, 8);
     gc_assert(!strncmp(signature, "SB.Crash", 8));
     gc_assert(read(fd, signature, 1) == 0);
     close(fd);
