@@ -34,16 +34,13 @@
     (declare (ignore x))
     (values t nil)))
 
-(defun find-named-callees (fun &key (type t) (name nil namep))
-  (let ((code (fun-code-header (%fun-fun fun))))
-    (loop for i from sb-vm:code-constants-offset below (code-header-words code)
+(defun find-named-callees (fun &key (name nil namep))
+  (sb-int:binding* ((code (fun-code-header (%fun-fun fun)))
+                    ((start count) (sb-kernel:code-header-fdefn-range code)))
+    (loop for i from start repeat count
           for c = (code-header-ref code i)
-          when (and (typep c 'sb-impl::fdefn)
-                    (let ((fun (sb-impl::fdefn-fun c)))
-                      (and (typep fun type)
-                           (or (not namep)
-                               (equal name (sb-impl::fdefn-name c))))))
-          collect (sb-impl::fdefn-fun c))))
+          when (or (not namep) (equal name (sb-kernel:fdefn-name c)))
+          collect (sb-kernel:fdefn-fun c))))
 
 (defun find-anonymous-callees (fun &key (type 'function))
   (let ((code (fun-code-header (%fun-fun fun))))
@@ -80,13 +77,14 @@
 (defun check-consing (yes/no form thunk times)
   (multiple-value-bind (before after)
       (collect-consing-stats thunk times)
-    (let ((consed-bytes (- after before)))
+    (let* ((consed-bytes (- after before))
+           (bytes-per-iteration (float (/ consed-bytes times))))
       (assert (funcall (if yes/no #'not #'identity)
-                       ;; I do not know why we do this comparasion,
-                       ;; the original code did, so I let it
-                       ;; in. Perhaps to prevent losage on GC
-                       ;; fluctuations, or something. --TCR.
-                       (< consed-bytes times))
+                       ;; If allocation really happened, it can't have been less than one cons cell
+                       ;; per iteration (unless the test is nondeterministic - but in that case
+                       ;; we can't really use this strategy anyway). So consider it to have consed
+                       ;; nothing if the fraction is too small.
+                       (< bytes-per-iteration (* 2 sb-vm:n-word-bytes)))
               ()
               "~@<Expected the form ~
                       ~4I~@:_~A ~0I~@:_~
@@ -94,7 +92,7 @@
                   ~D times resulted in the allocation of ~
                   ~D bytes~:[ (~,3F per run)~;~].~@:>"
               form yes/no times consed-bytes
-              (zerop consed-bytes) (float (/ consed-bytes times))))
+              (zerop consed-bytes) bytes-per-iteration))
     (values before after)))
 
 (defparameter +times+ 10000)
