@@ -498,11 +498,9 @@
 
 (test-util:with-test (:name :get-info-value-initializing
                       :skipped-on (not :sb-thread))
-  ;; Precompute random generalized function names for testing, some of which
-  ;; are "simple" (per the taxonomy of globaldb) and some hairy.
+  ;; Precompute random generalized function names for testing
   (let ((work (coerce (loop repeat 10000
-                            nconc (list `(setf ,(gensym)) ; simple name
-                                         (gensym))) ; very simple name
+                            nconc (list `(setf ,(gensym)) `(cas ,(gensym))))
                       'vector))
         (n-threads 10) readers writers fdefn-results random-results)
     (dotimes (i (ash n-threads -1))
@@ -598,8 +596,8 @@
                    for val = (info :variable :macro-expansion name)
                    do (assert (eql (or val 0) count))))))
     ;; Try it when names are symbols or "simple" 2-list names
-    (run (coerce (loop repeat 50
-                       for sym = (gensym)
+    (run (coerce (loop for sym = (gensym)
+                       repeat 50
                        nconc (list `(setf ,sym) sym))
                  'vector))
     ;; For hairy names, the tricky piece is in the rehash algorithm,
