@@ -59,7 +59,7 @@
     (pure-runner (pure-cload-files) 'cload-test log)
     (impure-runner (impure-load-files) 'load-test log)
     (impure-runner (impure-cload-files) 'cload-test log)
-    #-win32 (impure-runner (sh-files) 'sh-test log)
+    (impure-runner (sh-files) 'sh-test log)
     (log-file-elapsed-time "GRAND TOTAL" start-time log))
   (report)
   (sb-ext:exit :code (if (unexpected-failures) 1 104)))
@@ -196,12 +196,18 @@
   (flet ((maybe (p s) (and (find-package p)
                            (find-symbol s p))))
     `(sb-c::*code-serialno*
+      sb-c::*compile-elapsed-time*
+      sb-c::*compile-file-elapsed-time*
       sb-impl::*package-names-cookie*
       sb-impl::*available-buffers*
       sb-impl::*token-buf-pool*
       sb-impl::*user-hash-table-tests*
       sb-impl::**finalizer-store**
-      sb-vm::*codeblob-tree*
+      sb-impl::*pn-dir-cache*
+      sb-impl::*pn-cache*
+      sb-impl::*pn-cache-force-recount*
+      sb-vm::*immobile-codeblob-tree*
+      sb-vm::*dynspace-codeblob-tree*
       ,(maybe "SB-KERNEL" "*EVAL-CALLS*")
       sb-kernel::*type-cache-nonce*
       sb-ext:*gc-run-time*
@@ -217,6 +223,7 @@
       ,(maybe "SB-VM" "*STORE-BARRIERS-POTENTIALLY-EMITTED*")
       ,(maybe "SB-VM" "*STORE-BARRIERS-EMITTED*")
       ,(maybe "SB-INTERPRETER" "*LAST-TOPLEVEL-ENV*")
+      ,(maybe "SB-SYS" "*THRUPTION-PENDING*")
       sb-pcl::*dfun-constructors*
       #+win32 sb-impl::*waitable-timer-handle*
       #+win32 sb-impl::*timer-thread*)))
@@ -565,4 +572,23 @@
   (filter-test-files "*.impure-cload.lisp"))
 
 (defun sh-files ()
-  (filter-test-files "*.test.sh"))
+  (let ((result (filter-test-files "*.test.sh")))
+    #+unix result
+    ;; Rather than hack up the shell scripts which don't pass on #-unix
+    ;; (which would require at least a few lines of shell script and lisp
+    ;; to invoke SBCL and exit with some other code), just confine the kludge
+    ;; to this file.
+    #-unix
+    (if *explicit-test-files*
+        result
+      (remove-if
+       (lambda (x)
+         (member (pathname-name x)
+                 '("filesys.test" ; too many assertions about symlinks to care about just yet
+                   ;; foreign-test-noop-dlclose-test.c:1:10: fatal error: dlfcn.h: No such file or directory
+                   "foreign.test"
+                   ;; No built SBCL here (.../tests/run-sbcl-test-5863): run 'sh make.sh' first!
+                   "run-sbcl.test"
+                   "side-effectful-pathnames.test") ; no idea
+                 :test 'string=))
+        result))))
