@@ -35,6 +35,7 @@
            #:*scratch-file-prefix*
            #:with-scratch-file
            #:with-test-directory
+           #:generate-test-directory-name
            #:*test-directory*
            #:opaque-identity
            #:runtime #:split-string #:integer-sequence #:shuffle))
@@ -87,7 +88,6 @@
     ;; To actually exercise the type operation, remove the memoized parse.
     (sb-int:drop-all-hash-caches)
     (let ((b (sb-kernel:values-specifier-type right)))
-      (assert (not (eq a b)))
       (sb-kernel:type= a b))))
 
 (defmacro assert-tri-eq (expected-result expected-certainp form)
@@ -238,17 +238,18 @@
 ;;; but the nice side effect is that the tests finish quicker.
 (defmacro with-test ((&key fails-on broken-on skipped-on name serial slow)
                      &body body)
-  (flet ((name-ok (x y)
-           (declare (ignore y))
-           (typecase x
-             (symbol (let ((package (symbol-package x)))
-                       (or (null package)
-                           (sb-int:system-package-p package)
-                           (eql package (find-package "CL"))
-                           (eql package (find-package "KEYWORD")))))
-             (integer t))))
-    (unless (tree-equal name name :test #'name-ok)
-      (error "test name must be all-keywords: ~S" name)))
+  ;; Failing and skipped tests are written into a summary file which is later read back.
+  ;; To guarantee readability there can't be symbols in random packages.
+  (setq name (sb-int:named-let ensure-ok ((x name))
+               (etypecase x
+                 (cons (cons (ensure-ok (car x)) (ensure-ok (cdr x))))
+                 (symbol (let ((package (symbol-package x)))
+                           (if (or (null package)
+                                   (sb-int:system-package-p package)
+                                   (eql package (find-package "CL"))
+                                   (eql package (find-package "KEYWORD")))
+                               x (copy-symbol x))))
+                 (integer x))))
   (cond
     ((broken-p broken-on)
      `(progn
@@ -888,14 +889,22 @@
 
 (defvar *test-directory*)
 
+(defun generate-test-directory-name ()
+  ;; Why aren't we using TMPDIR???
+  (merge-pathnames
+   (make-pathname :directory `(:relative ,(write-to-string (sb-unix:unix-getpid))))
+   (parse-native-namestring (posix-getenv "TEST_DIRECTORY")
+                            nil *default-pathname-defaults*
+                            :as-directory t)))
+
 (defun call-with-test-directory (fn)
   ;; FIXME: this writes into the source directory depending on whether
   ;; TEST_DIRECTORY has been made to point elsewhere or not.
-  (let ((test-directory (parse-native-namestring (posix-getenv "TEST_DIRECTORY")
-                                                 nil *default-pathname-defaults*
-                                                 :as-directory t)))
+  (let ((test-directory (generate-test-directory-name)))
     (ensure-directories-exist test-directory)
     (unwind-protect
+         ;; WHY REBIND *DEFAULT-PATHNAME-DEFAULTS* ? THIS SUCKS!
+         ;; (It means we can't use the WITH-TEST-DIRECTORY macro for most things)
          (let ((*default-pathname-defaults* test-directory)
                (*test-directory* test-directory))
            (funcall fn test-directory))
@@ -965,3 +974,8 @@
   (when (find-package "SB-SPROF")
     (format t "INFO: disabling SB-SPROF~%")
     (funcall (intern "STOP-PROFILING" "SB-SPROF"))))
+
+;;; This unexported symbol emulates SB-RT. Please don't use it in new tests
+(defmacro deftest (name form &rest results) ; use SB-RT syntax
+  `(test-util:with-test (:name ,(sb-int:keywordicate name))
+     (assert (equalp (multiple-value-list ,form) ',results))))
