@@ -299,6 +299,14 @@
 (define-condition condition-foo2 (condition-foo1) ())
 (define-condition condition-foo3 (condition-foo2) ())
 (define-condition condition-foo4 (condition-foo3) ())
+(with-test (:name :add-subclassoid)
+  (flet ((has-subs (name n)
+           (= n (length (sb-kernel:classoid-subclasses
+                         (sb-kernel:find-classoid name))))))
+  (assert (has-subs 'condition-foo1 3)) ; has foo{2,3,4}
+  (assert (has-subs 'condition-foo2 2)) ; has foo{3,4}
+  (assert (has-subs 'condition-foo3 1)) ; has foo4
+  (assert (has-subs 'condition-foo4 0))))
 
 ;;; inline type tests
 (format t "~&/setting up *TESTS-OF-INLINE-TYPE-TESTS*~%")
@@ -928,7 +936,8 @@
   (let* ((hair (sb-kernel:specifier-type '(sb-kernel:unboxed-array 1)))
          (xform (sb-c::source-transform-union-typep 'myobj hair)))
     (assert (equal xform
-                   '(or (typep myobj '(and vector (not (array t))))))))
+                   '(or (typep myobj
+                               '(and vector (not (array t)) (not (array nil))))))))
 
   ;; Exclude one subtype at a time and make sure they all work.
   (dotimes (i (length sb-vm:*specialized-array-element-type-properties*))
@@ -1022,3 +1031,20 @@
     (assert-tri-eq nil nil (sb-kernel:type= (sb-kernel:specifier-type 'float) unk))
     (assert-tri-eq nil nil (sb-kernel:type= (sb-kernel:specifier-type 'pathname) unk))
     (assert-tri-eq nil nil (sb-kernel:type= (sb-kernel:specifier-type 'sequence) unk))))
+
+(with-test (:name :lp-308938) ; got silently fixed in git rev ef8c95377a55
+  (multiple-value-bind (answer certain)
+      (subtypep '(or (satisfies x) string)
+                '(or (satisfies x) integer))
+    (assert (and (not answer) (not certain))))
+  (multiple-value-bind (answer certain)
+      (subtypep 'string '(or (satisfies x) integer))
+    (assert (and (not answer) (not certain)))))
+
+(deftype jn-even () '(and integer (or (eql 0) (satisfies f))))
+(deftype jn-odd () '(and integer (or (eql 1) (satisfies g))))
+(with-test (:name :lp-1528837) ; probably the same as the preceding fix
+  (multiple-value-bind (answer certain) (subtypep 'jn-odd 'jn-even)
+    (assert (and (not answer) (not certain))))
+  (multiple-value-bind (answer certain) (subtypep 'jn-even 'jn-odd)
+    (assert (and (not answer) (not certain)))))
