@@ -15,17 +15,15 @@
   (flet ((try (f) (assert-error (funcall f 'hash-table 3))))
     (mapc #'try '(typexpand-1 typexpand typexpand-all))))
 
-(with-test (:name :stream-layout-bits)
-  (loop for wrapper being each hash-value
-        of (sb-kernel:classoid-subclasses (sb-kernel:find-classoid 't))
-        do (flet ((check-bit (bit ancestor-type)
-                    (let ((ancestor (sb-kernel:find-layout ancestor-type)))
-                      (when (or (eq wrapper ancestor)
-                                (find ancestor (sb-kernel:wrapper-inherits wrapper)))
-                        (assert (logtest bit (sb-kernel:wrapper-flags wrapper)))))))
-              (check-bit sb-kernel:+stream-layout-flag+ 'stream)
-              (check-bit sb-kernel:+string-stream-layout-flag+ 'string-stream)
-              (check-bit sb-kernel:+file-stream-layout-flag+ 'file-stream))))
+(with-test (:name :no-*-as-t) ; lp#1860919
+  (assert-signal (sb-kernel:specifier-type '(function (*) t)) warning)
+  (dolist (f '((lambda (x) (the * x))
+               (lambda (x) (declare (* x)) x)
+               (lambda (x) (declare (type * x)) x)))
+    (multiple-value-bind (f warn err)
+        (let ((*error-output* (make-broadcast-stream))) (compile nil f))
+      (declare (ignore f))
+      (assert (and warn err)))))
 
 (with-test (:name (typep sb-kernel:ctypep))
   (locally
@@ -169,7 +167,7 @@
   (assert-tri-eq t   t (subtypep '(function ()) '(function (&rest t))))
   (assert-tri-eq nil t (subtypep '(function (&rest t)) '(function ())))
   (assert-tri-eq t   t (subtypep '(function)
-                                 '(function (&optional * &rest t))))
+                                 '(function (&optional t &rest t))))
   (assert-tri-eq nil t (subtypep '(function) '(function (t &rest t))))
   (assert-tri-eq t   t (subtypep 'function '(function)))
   (assert-tri-eq t   t (subtypep '(function) 'function)))
@@ -638,7 +636,7 @@
                (values bit &optional *)
                (values bit &rest *)
                (values bit &rest *)))
-    (assert-error (sb-kernel:values-specifier-type x))))
+    (assert-signal (sb-kernel:values-specifier-type x) warning)))
 
 (with-test (:name :classoids-as-type-specifiers)
   (dolist (classoid (list (find-classoid 'integer)
@@ -672,10 +670,42 @@
             (assert (eq (sb-int:info :type :kind s) :primitive))
             (assert (eq (sb-int:info :type :kind s) :instance)))))))
 
-(with-test (:name :make-numeric-type)
+(with-test (:name (make-numeric-type :smoke))
   (assert (eq (make-numeric-type :class 'integer :low '(4) :high '(5))
               *empty-type*)))
 
+(with-test (:name (make-numeric-type :union))
+  (assert (equal (type-specifier (make-numeric-type :low '(-79106810381456307)))
+                 `(or (rational (-79106810381456307))
+                      (single-float (-7.910681e16))
+                      (double-float (-7.91068103814563d16))))))
+
+(with-test (:name (make-numeric-type :infinities))
+  ;; Without class
+  (assert (equal (type-specifier
+                  (make-numeric-type :low sb-ext:single-float-negative-infinity
+                                     :high sb-ext:single-float-negative-infinity))
+                 `(or (single-float ,sb-ext:single-float-negative-infinity
+                                    ,sb-ext:single-float-negative-infinity)
+                      (double-float ,sb-ext:double-float-negative-infinity
+                                    ,sb-ext:double-float-negative-infinity))))
+  (assert (equal (type-specifier
+                  (make-numeric-type :low sb-ext:single-float-negative-infinity))
+                 'real))
+  ;; With FLOAT class
+  (assert (equal (type-specifier
+                  (make-numeric-type :class 'float
+                                     :low sb-ext:single-float-negative-infinity
+                                     :high sb-ext:single-float-negative-infinity))
+                 `(or (single-float ,sb-ext:single-float-negative-infinity
+                                    ,sb-ext:single-float-negative-infinity)
+                      (double-float ,sb-ext:double-float-negative-infinity
+                                    ,sb-ext:double-float-negative-infinity))))
+  (assert (equal (type-specifier
+                  (make-numeric-type :class 'float
+                                     :low sb-ext:single-float-negative-infinity))
+                 `float)))
+
 (with-test (:name :prettier-union-types :skipped-on (not :sb-unicode))
   ;; (OR STRING BIGNUM) used to unparse as
   ;; (OR (VECTOR CHARACTER) BASE-STRING (INTEGER * -4611686018427387905)
@@ -736,16 +766,10 @@
   (assert (eq (type-of (eval '(sb-kernel:%make-instance 12)))
               'sb-kernel:instance)))
 
-(with-test (:name :make-numeric-type-union)
-  (assert (equal (sb-kernel:type-specifier
-                  (sb-kernel:make-numeric-type :low '(-79106810381456307)))
-                 '(or (double-float (-7.91068103814563d16)) (single-float (-7.910681e16))
-                   (rational (-79106810381456307))))))
-
 (with-test (:name (:cons-union :lp1912863))
   (let ((c (cons 2 4)))
     (assert (not (typep c '(or (cons (integer 0 8) (integer 5 15))
-                               (cons (integer 3 15) (integer 4 14))))))))
+                            (cons (integer 3 15) (integer 4 14))))))))
 
 (with-test (:name (:rational-union :equivalent-to-t))
   (let ((type '(or (integer * -1) (rational -1/2 1/2) (integer 1) (not integer))))
@@ -793,3 +817,19 @@
 (with-test (:name (:lp1916233))
   (assert-tri-eq t t (subtypep '(cons (or (simple-array ratio) simple-array) nil) nil))
   (assert-tri-eq t t (subtypep '(or (array ratio) sequence) t)))
+
+(defun my-widetag-of (x)
+  (sb-sys:sap-ref-8 (sb-sys:int-sap (sb-kernel:get-lisp-obj-address x))
+                    (- sb-vm:other-pointer-lowtag)))
+;;; I'll bet that nothing anywhere tests this
+(with-test (:name :nil-has-symbol-widetag
+            :skipped-on (:or :ppc64 :big-endian))
+  (assert (= (my-widetag-of nil) (my-widetag-of t))))
+
+(with-test (:name :array-rank-deriver-negation)
+  (assert
+   (ctype= (caddr
+            (sb-kernel:%simple-fun-type
+             (checked-compile
+              `(lambda (a) (array-rank (the (not (array t)) a))))))
+           `(values array-rank &optional))))
