@@ -9,7 +9,7 @@
 ;;;; absolutely no warranty. See the COPYING and CREDITS files for
 ;;;; more information.
 
-#-x86-64 (sb-ext:exit :code 104)
+#-x86-64 (invoke-restart 'run-tests::skip-file)
 
 ;;; This trivial function failed to compile due to rev 88d078fe
 (defun foo (&key k)
@@ -182,22 +182,6 @@
       (assert (search "; #<SB-KERNEL:WRAPPER " line))
       (assert (search " SB-ASSEM:LABEL" line)))))
 
-#+immobile-code
-(with-test (:name :reference-assembly-tramp)
-  (dolist (testcase '(("FUNCALLABLE-INSTANCE-TRAMP"
-                       sb-kernel:%make-funcallable-instance)))
-    (let ((lines
-           (split-string
-            (with-output-to-string (stream)
-              (let ((sb-disassem:*disassem-location-column-width* 0))
-                (disassemble (cadr testcase) :stream stream)))
-            #\newline)))
-      (assert (loop for line in lines
-                    thereis (and (search "LEA" line)
-                                 (search "RIP" line) ; require RIP-relative mode
-                                 ;; and verify disassembly
-                                 (search (car testcase) line)))))))
-
 #+immobile-code ; uses SB-C::*COMPILE-TO-MEMORY-SPACE*
 (with-test (:name :static-link-compile-to-memory)
   (let* ((string
@@ -215,40 +199,7 @@
     (assert (find-line "CALL" "FUNCTION GENSYM"))
     (assert (find-line "JMP" "FUNCTION PRINT")))))
 
-#+immobile-code
-(with-test (:name :static-unlinker)
-  (let ((sb-c::*compile-to-memory-space* :immobile))
-    (declare (muffle-conditions style-warning))
-    (flet ((disassembly-lines (name)
-             (split-string
-              (with-output-to-string (s)
-                (let ((sb-disassem:*disassem-location-column-width* 0))
-                  (disassemble name :stream s)))
-              #\newline))
-           (expect (match lines)
-             (assert (loop for line in lines
-                           thereis (search match line)))))
-      (compile 'h '(lambda (x) (1+ x)))
-      (setf (symbol-function 'g) #'h (symbol-function 'f) #'h)
-      (compile 'c '(lambda (x) (g x)))
-      (compile 'd '(lambda (x) (f (g x))))
-      ;; The FDEFN-FUN of F is same as that of G.
-      ;; Statically linking D should not patch the fdefn calls into static calls
-      ;; because it can't unambiguously be undone without storing additional data
-      ;; about where patches were performed to begin with.
-      (sb-vm::statically-link-core :callers '(c d))
-      (let ((lines (disassembly-lines 'c)))
-        (expect "#<FUNCTION H>" lines))
-      (let ((lines (disassembly-lines 'd)))
-        (expect "#<FUNCTION H>" lines))
-      (setf (symbol-function 'g) #'+)
-      (let ((lines (disassembly-lines 'c)))
-        (expect "#<FDEFN G>" lines))
-      (let ((lines (disassembly-lines 'd)))
-        (expect "#<FDEFN G>" lines)
-        (expect "#<FUNCTION H>" lines)))))
-
-(with-test (:name :c-call :skipped-on (and :win32 :x86-64))
+(with-test (:name :c-call :skipped-on :win32)
   (let* ((lines (split-string
                  (with-output-to-string (s)
                    (let ((sb-disassem:*disassem-location-column-width* 0))
@@ -257,7 +208,7 @@
          (c-call (find "os_deallocate" lines :test #'search)))
     ;; Depending on #+immobile-code it's either direct or memory indirect.
     #+immobile-code (assert (search "CALL #x" c-call))
-    #-immobile-code (assert (search "CALL QWORD PTR [#x" c-call))))
+    #-immobile-code (assert (search "CALL [#x" c-call))))
 
 (with-test (:name :set-symbol-value-imm)
   (let (success)
@@ -403,7 +354,7 @@
   (assert (thing-ref-thing-ref (make-thing :x (make-thing :x 3)) 'foo)))
 
 (with-test (:name :huge-code :skipped-on (not :immobile-code))
-  (sb-vm::allocate-code-object :immobile 0 4 (* 2 1024 1024)))
+  (sb-vm::allocate-code-object :immobile 4 (* 2 1024 1024)))
 
 (defun bbb (x y z)
   ;; I don't want the number of expected comparisons to depend on whether
@@ -763,6 +714,7 @@
                           (search "CMP DWORD PTR" line)))))
 
 (with-test (:name :thread-local-unbound)
+  (declare (optimize safety))
   (let ((c (nth-value 1 (ignore-errors sb-c::*compilation*))))
     (assert (eq (cell-error-name c) 'sb-c::*compilation*))))
 
@@ -814,11 +766,6 @@
                          lines)
                1))))
 
-#+compact-instance-header
-(with-test (:name :gf-self-contained-trampoline)
-  (let ((l (sb-kernel:find-layout 'standard-generic-function)))
-    (assert (/= (sb-kernel:wrapper-bitmap l) sb-kernel:+layout-all-tagged+))))
-
 (with-test (:name :known-array-rank)
   (flet ((try (type)
            (let ((lines
@@ -871,12 +818,10 @@
   (check-arrayp-cmp-opcodes 1 'simple-base-string)
   #+sb-unicode
   (check-arrayp-cmp-opcodes 1 'sb-kernel::simple-character-string)
-  ;; FIXME: (AND STRING (NOT SIMPLE-STRING)) executs 4 tests
-  ;; but it denotes the same set of objects as (AND STRING (NOT SIMPLE-ARRAY)).
-  ;; The problem is with type algebra, not in the backend.
-
-  ;; FIXME: this should be 1 comparison: widetag >= start-of-complex-widetags
-  (check-arrayp-cmp-opcodes 2 '(and array (not simple-array)))
+  (check-arrayp-cmp-opcodes 1 '(and string (not simple-array)))
+  (check-arrayp-cmp-opcodes #+sb-unicode 1
+                            #-sb-unicode 2
+                            '(and array (not simple-array)))
 
   ;; some other interesting pairs
   ;; This was passing just by random coincidence.
@@ -904,7 +849,24 @@
   ;; and some range-based testing rather than just 2 EQ tests.
   (check-integerp-cmp-opcodes 2 '(integer 1 2)))
 
-(defun typep-asm-code-length (type)
+(defun show-pretty-lines (type lines)
+  (format t ";;;; Type: ~s~%" type)
+  (dolist (line lines)
+    (when (plusp (length line))
+      (let* ((label
+              (if (char= (char line 0) #\L)
+                  (subseq line 0 (1+ (position #\: line)))
+                  ""))
+             (opcode-bytes
+              (progn
+                (setq line (string-left-trim " " (subseq line (length label))))
+                (subseq line 0 (position #\Space line)))))
+        (setq line (string-left-trim " " (subseq line (length opcode-bytes))))
+        (format t "; ~4a~20a~a~%"
+                label opcode-bytes line))))
+  (terpri))
+
+(defun typep-asm-code-length (type &optional print)
   (let* ((lines
           (disassembly-lines
            (compile nil
@@ -915,6 +877,7 @@
           (some (lambda (x) (or (search "#<FUNCTION" x)
                                 (search "#<FDEFN" x)))
                 lines)))
+    (when print (show-pretty-lines type lines))
     (values (length lines) callp)))
 
 ;;; Counting instructions of assembly is sort of a very rough guess
@@ -940,7 +903,48 @@
 (write-golden-typep-data "../interesting-types.lisp-expr"
                          "typep-golden-data.txt")
 
-(defun compare-to-golden-typep-data (pathname)
+;;; FIXME: Some time after the MANY-INTERESTING-ARRAY-TYPES test
+;;; got disabled, the code size changed either due to a regression,
+;;; or more aggressive inlining of ">=" and "<="
+;;; which now causes these warnings. Figure out the new baseline.
+;;
+;; WARNING: (AND INTEGER (NOT (SIGNED-BYTE 31))) was 32 is 38
+;; WARNING: (AND INTEGER (NOT (SIGNED-BYTE 32))) was 32 is 38
+;; WARNING: (INTEGER (0) *) was 18 is 22
+;; WARNING: (INTEGER * -1) was 18 is 22
+;; WARNING: (INTEGER * -1000000) was 18 is 22
+;; WARNING: (INTEGER * -1073741825) was 18 is 23
+;; WARNING: (INTEGER * -12) was 18 is 22
+;; WARNING: (INTEGER * -14) was 18 is 22
+;; WARNING: (INTEGER * -17) was 18 is 22
+;; WARNING: (INTEGER * -2) was 18 is 22
+;; WARNING: (INTEGER * -2147483649) was 18 is 23
+;; WARNING: (INTEGER * -4611686018427387901) was 18 is 23
+;; WARNING: (INTEGER * -54043195528445951) was 18 is 23
+;; WARNING: (INTEGER * -541073411) was 18 is 22
+;; WARNING: (INTEGER * -6) was 18 is 22
+;; WARNING: (INTEGER * -9) was 18 is 22
+;; WARNING: (INTEGER * 0) was 24 is 22
+;; WARNING: (INTEGER * 20) was 18 is 22
+;; WARNING: (INTEGER * 2047) was 18 is 22
+;; WARNING: (INTEGER * 2147483647) was 18 is 23
+;; WARNING: (INTEGER * 4611686018427387900) was 18 is 23
+;; WARNING: (INTEGER * 576460752303423487) was 18 is 23
+;; WARNING: (INTEGER * 65535) was 18 is 22
+;; WARNING: (INTEGER -1) was 18 is 22
+;; WARNING: (INTEGER 1) was 18 is 22
+;; WARNING: (INTEGER 1073741824) was 18 is 23
+;; WARNING: (INTEGER 1152921504606846961) was 18 is 23
+;; WARNING: (INTEGER 1899) was 18 is 22
+;; WARNING: (INTEGER 2) was 18 is 22
+;; WARNING: (INTEGER 2147483648) was 18 is 23
+;; WARNING: (INTEGER 2305843009214) was 18 is 23
+;; WARNING: (INTEGER 4611686018427387900) was 18 is 23
+;; WARNING: (INTEGER 4611686018427387901) was 18 is 23
+;; WARNING: (INTEGER 4611686018427387902) was 18 is 23
+;; WARNING: (INTEGER 4611686018427387903) was 18 is 23
+;; WARNING: (INTEGER 63) was 18 is 22
+(defun compare-to-golden-typep-data (pathname &optional print)
   (with-open-file (input pathname)
     (let ((*package* (find-package "SB-KERNEL")))
       (loop (let ((line (read-line input nil input)))
@@ -948,13 +952,15 @@
               (with-input-from-string (stream line :start 2)
                 (let ((expect-n (read stream))
                       (type (read stream)))
-                  (multiple-value-bind (linecount callp) (typep-asm-code-length type)
+                  (multiple-value-bind (linecount callp)
+                      (typep-asm-code-length type print)
                     (declare (ignore callp))
                     (when (/= linecount expect-n)
                       (warn "~S was ~d is ~d" type expect-n linecount))))))))))
 
-(with-test (:name :many-interesting-array-types :skipped-on (or (:not :sb-unicode)
-                                                                (not :sb-devel)))
+(with-test (:name :many-interesting-array-types
+                  :skipped-on (:or (:not :sb-unicode)
+                                   (:not :immobile-space)))
   (compare-to-golden-typep-data "typep-golden-data.txt"))
 
 (with-test (:name :integerp->bignump-strength-reduction)
@@ -971,6 +977,7 @@
   ;; Previously there would have been one from BOUNDP and one from SYMBOL-VALUE.
   (let ((lines (disassembly-lines
                 '(lambda ()
+                  (declare (optimize safety))
                   (if (boundp 'sb-c::*compilation*) sb-c::*compilation*) '(hi)))))
     (dolist (line lines)
       (assert (not (search "ERROR" line))))
@@ -1176,3 +1183,122 @@
 (with-test (:name :disassemble-symbol-global-value)
   (assert (loop for line in (disassembly-lines '(lambda () *myglobalvar*))
                 thereis (search "*MYGLOBALVAR*" line))))
+
+(defun compiler-trace-output-lines (lexpr)
+  (let ((string-stream (make-string-output-stream)))
+    (let ((sb-c::*compiler-trace-output* string-stream)
+          (sb-c::*compile-trace-targets* '(:vop))
+          (*print-pretty* nil))
+      (compile nil lexpr))
+    (split-string (get-output-stream-string string-stream)
+                  #\newline)))
+
+(defun assert-has-gc-barrier (match-string lines &optional (skip 0))
+
+  (let (found)
+;; Look for something like this in the trace output:
+;; "VOP CLOSURE-INIT t9[RBX(d)] :NORMAL t13[RSI(d)] :NORMAL {0} "
+;; "        MOV     0, #<TN t14[RAX(u)] :NORMAL>, #<TN t9[RBX(d)] :NORMAL>"
+;; "        SHR     0, #<TN t14[RAX(u)] :NORMAL>, 10"
+;; "        AND     6, #<TN t14[RAX(u)] :NORMAL>, #S(FIXUP :NAME NIL :FLAVOR GC-BARRIER :OFFSET 0)"
+;; "        MOV     4, PTR [R12+RAX+0], 0"
+;; "        MOV     7, PTR [RBX+5], #<TN t13[RSI(d)] :NORMAL>"
+    (loop while lines
+          do
+          (when (and (search match-string (car lines))
+                     (minusp (decf skip)))
+            (setq found t)
+            (let ((barrier-fixup-line (nth 3 lines)))
+              (assert (search ":FLAVOR GC-BARRIER" barrier-fixup-line))
+              (return)))
+          (pop lines))
+    ;; Fail if we didn't find the sentinel CLOSURE-INIT
+    (assert found)))
+
+(with-test (:name :closure-init-gc-barrier)
+  (let ((lines
+         (compiler-trace-output-lines
+           '(lambda (address)
+             (declare (sb-vm:word address))
+             (let ((sap (sb-sys:int-sap address)))
+               (lambda () sap))))))
+    (assert-has-gc-barrier "CLOSURE-INIT" lines)))
+
+(defstruct (point (:constructor make-point (x))) x (y 0) (z 0))
+(with-test (:name :structure-init-gc-barrier)
+  (let ((lines
+          (compiler-trace-output-lines
+           '(lambda (val)
+             (declare (double-float val) (inline make-point))
+             (let ((neg (- val))) (make-point neg))))))
+    (assert-has-gc-barrier "SET-SLOT" lines
+                           #-compact-instance-header 1)))
+
+(with-test (:name :system-tlabs)
+  (when (find-symbol "SYS-ALLOC-TRAMP" "SB-VM")
+    (assert (loop for line in (disassembly-lines 'sb-impl:test-make-packed-info)
+                  thereis (search "SYS-ALLOC-TRAMP" line)))
+    (assert (loop for line in (disassembly-lines 'sb-impl:test-copy-packed-info)
+                  thereis (search "SYS-ALLOC-TRAMP" line)))
+    (let ((f (compile nil '(lambda (x)
+                            (declare (sb-c::tlab :system))
+                            (sb-pcl::%copy-cache x)))))
+      (assert (loop for line in (disassembly-lines f)
+                    thereis (search "SYS-ALLOC-TRAMP" line))))))
+
+(defun find-in-disassembly (string lambda-expression)
+  (let ((disassembly
+         (with-output-to-string (s)
+           (disassemble (compile nil lambda-expression) :stream s))))
+    (loop for line in (split-string disassembly #\Newline)
+          thereis (search string line))))
+
+#+immobile-space
+(with-test (:name :disassemble-alien-linkage-table-ref
+                  :fails-on (not :sb-thread))
+  (dolist (memspace '(:dynamic :immobile))
+    (let ((sb-c::*compile-to-memory-space* memspace))
+      (assert (find-in-disassembly
+               (if (eq sb-c::*compile-to-memory-space* :immobile) "lose" "&lose")
+               '(lambda ()
+                 (declare (optimize (sb-c::alien-funcall-saves-fp-and-pc 0)))
+                 (alien-funcall (extern-alien "lose" (function void))))))
+      (assert (find-in-disassembly
+               "&verify_gens"
+               '(lambda ()
+                 (extern-alien "verify_gens" char)))))))
+
+;;; This tests that the x86-64 disasembler does not crash
+;;; on LEA with a rip-relative operand and no label.
+(with-test (:name (disassemble :no-labels))
+  (let* ((lines
+          (split-string
+           (with-output-to-string (stream)
+             ;; A smallish function whose code happens to contain
+             ;; the thing under test.
+             (disassemble 'sb-impl::inspector :stream stream))
+          #\Newline))
+         (line (find "; = L0" lines :test 'search)))
+    (assert (search "LEA " line)) ; verify our test precondition
+    ;; Now just disassemble without labels and see that we don't crash
+    (disassemble 'sb-impl::inspector
+                 :use-labels nil
+                 :stream (make-broadcast-stream))))
+
+#+immobile-space
+(with-test (:name (disassemble :static-call))
+  (dolist (sym '(sb-kernel:two-arg-* sb-kernel:two-arg-/
+                 sb-kernel:two-arg-+ sb-kernel:two-arg--))
+    (let ((f (let ((sb-c:*compile-to-memory-space* :dynamic))
+               (compile nil `(lambda (x y) (,sym x y))))))
+      (assert (loop for line in (disassembly-lines f)
+                    thereis (search (princ-to-string sym) line))))))
+
+(with-test (:name :closure-gc-barrier)
+  (let ((lines
+          (compiler-trace-output-lines
+           '(lambda ()
+             (let ((c (cons 1 2)))
+               (lambda (x)
+                 (setf (car c) x)))))))
+    (assert-has-gc-barrier "SET-SLOT" lines)))
